# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os

from powerline.lib.unicode import out_u
from powerline.theme import requires_segment_info
from powerline.segments import Segment, with_docstring


@requires_segment_info
def environment(pl, segment_info, variable=None):
	'''Return the value of any defined environment variable

	:param string variable:
		The environment variable to return if found
	'''
	return segment_info['environ'].get(variable, None)


@requires_segment_info
def virtualenv(pl, segment_info, ignore_venv=False, ignore_conda=False, ignored_names=("venv", ".venv")):
	'''Return the name of the current Python or conda virtualenv.
	:param list ignored_names:
		Names of venvs to ignore. Will then get the name of the venv by ascending to the parent directory
	:param bool ignore_venv:
		Whether to ignore virtual environments. Default is False.
	:param bool ignore_conda:
		Whether to ignore conda environments. Default is False.
	'''
	if not ignore_venv:
		for candidate in reversed(segment_info['environ'].get('VIRTUAL_ENV', '').split("/")):
			if candidate and candidate not in ignored_names:
				return candidate
	if not ignore_conda:
		for candidate in reversed(segment_info['environ'].get('CONDA_DEFAULT_ENV', '').split("/")):
			if candidate and candidate not in ignored_names:
				return candidate
	return None


@requires_segment_info
class CwdSegment(Segment):
	def argspecobjs(self):
		for obj in super(CwdSegment, self).argspecobjs():
			yield obj
		yield 'get_shortened_path', self.get_shortened_path

	def omitted_args(self, name, method):
		if method is self.get_shortened_path:
			return ()
		else:
			return super(CwdSegment, self).omitted_args(name, method)

	def get_shortened_path(self, pl, segment_info, shorten_home=True, **kwargs):
		try:
			path = out_u(segment_info['getcwd']())
		except OSError as e:
			if e.errno == 2:
				# user most probably deleted the directory
				# this happens when removing files from Mercurial repos for example
				pl.warn('Current directory not found')
				return '[not found]'
			else:
				raise
		if shorten_home:
			home = segment_info['home']
			if home:
				home = out_u(home)
				if path.startswith(home):
					path = '~' + path[len(home):]
		return path

	def __call__(self, pl, segment_info,
	             dir_shorten_len=None,
	             dir_limit_depth=None,
	             use_path_separator=False,
	             ellipsis='...',
	             **kwargs):
		cwd = self.get_shortened_path(pl, segment_info, **kwargs)
		cwd_split = cwd.split(os.sep)
		cwd_split_len = len(cwd_split)
		cwd = [i[0:dir_shorten_len] if dir_shorten_len and i else i for i in cwd_split[:-1]] + [cwd_split[-1]]
		if dir_limit_depth and cwd_split_len > dir_limit_depth + 1:
			del(cwd[0:-dir_limit_depth])
			if ellipsis is not None:
				cwd.insert(0, ellipsis)
		ret = []
		if not cwd[0]:
			cwd[0] = '/'
		draw_inner_divider = not use_path_separator
		for part in cwd:
			if not part:
				continue
			if use_path_separator:
				part += os.sep
			ret.append({
				'contents': part,
				'divider_highlight_group': 'cwd:divider',
				'draw_inner_divider': draw_inner_divider,
			})
		ret[-1]['highlight_groups'] = ['cwd:current_folder', 'cwd']
		if use_path_separator:
			ret[-1]['contents'] = ret[-1]['contents'][:-1]
			if len(ret) > 1 and ret[0]['contents'][0] == os.sep:
				ret[0]['contents'] = ret[0]['contents'][1:]
		return ret


cwd = with_docstring(CwdSegment(),
'''Return the current working directory.

Returns a segment list to create a breadcrumb-like effect.

:param int dir_shorten_len:
	shorten parent directory names to this length (e.g. 
	:file:`/long/path/to/powerline` → :file:`/l/p/t/powerline`)
:param int dir_limit_depth:
	limit directory depth to this number (e.g. 
	:file:`/long/path/to/powerline` → :file:`⋯/to/powerline`)
:param bool use_path_separator:
	Use path separator in place of soft divider.
:param bool shorten_home:
	Shorten home directory to ``~``.
:param str ellipsis:
	Specifies what to use in place of omitted directories. Use None to not 
	show this subsegment at all.

Divider highlight group used: ``cwd:divider``.

Highlight groups used: ``cwd:current_folder`` or ``cwd``. It is recommended to define all highlight groups.
''')


try:
	import psutil

	# psutil-2.0.0: psutil.Process.username is unbound method
	if callable(psutil.Process.username):
		def _get_user():
			return psutil.Process(os.getpid()).username()
	# pre psutil-2.0.0: psutil.Process.username has type property
	else:
		def _get_user():
			return psutil.Process(os.getpid()).username
except ImportError:
	try:
		import pwd
	except ImportError:
		from getpass import getuser as _get_user
	else:
		try:
			from os import geteuid as getuid
		except ImportError:
			from os import getuid

		def _get_user():
			return pwd.getpwuid(getuid()).pw_name


username = False
# os.geteuid is not available on windows
_geteuid = getattr(os, 'geteuid', lambda: 1)


@requires_segment_info
def user(pl, segment_info, hide_user=None, hide_domain=False):
	'''Return the current user.

	:param str hide_user:
		Omit showing segment for users with names equal to this string.
	:param bool hide_domain:
		Drop domain component if it exists in a username (delimited by '@').

	Highlights the user with the ``superuser`` if the effective user ID is 0.

	Highlight groups used: ``superuser`` or ``user``. It is recommended to define all highlight groups.
	'''
	global username
	if (
		segment_info['environ'].get('_POWERLINE_RUNNING_SHELL_TESTS')
		== 'ee5bcdc6-b749-11e7-9456-50465d597777'
	):
		return 'user'
	if username is False:
		username = _get_user()
	if username is None:
		pl.warn('Failed to get username')
		return None
	if username == hide_user:
		return None
	if hide_domain:
		try:
			username = username[:username.index('@')]
		except ValueError:
			pass
	euid = _geteuid()
	return [{
		'contents': username,
		'highlight_groups': ['user'] if euid != 0 else ['superuser', 'user'],
	}]



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import re

from powerline.theme import requires_segment_info
from powerline.bindings.wm import get_i3_connection

WORKSPACE_REGEX = re.compile(r'^[0-9]+: ?')

def workspace_groups(w):
	group = []
	if w.focused:
		group.append('workspace:focused')
		group.append('w_focused')
	if w.urgent:
		group.append('workspace:urgent')
		group.append('w_urgent')
	if w.visible:
		group.append('workspace:visible')
		group.append('w_visible')
	group.append('workspace')
	return group


def format_name(name, strip=False):
	if strip:
		return WORKSPACE_REGEX.sub('', name, count=1)
	return name


def is_empty_workspace(workspace, containers):
	if workspace.focused or workspace.visible:
		return False
	wins = [win for win in containers[workspace.name].leaves()]
	return False if len(wins) > 0 else True

WS_ICONS = {"multiple": "M"}

def get_icon(workspace, separator, icons, show_multiple_icons, ws_containers):
	icons_tmp = WS_ICONS
	icons_tmp.update(icons)
	icons = icons_tmp

	wins = [win for win in ws_containers[workspace.name].leaves() \
		if win.parent.scratchpad_state == 'none']
	if len(wins) == 0:
		return ''

	result = ''
	cnt = 0
	for key in icons:
		if not icons[key] or len(icons[key]) < 1:
			continue
		if any(key in win.window_class for win in wins if win.window_class):
			result += (separator if cnt > 0 else '') + icons[key]
			cnt += 1
	if not show_multiple_icons and cnt > 1:
		if 'multiple' in icons:
			return icons['multiple']
		else:
			return ''
	return result

@requires_segment_info
def workspaces(pl, segment_info, only_show=None, output=None, strip=0, format='{name}',
	icons=WS_ICONS, sort_workspaces=False, show_output=False, priority_workspaces=[],
	hide_empty_workspaces=False):
	'''Return list of used workspaces

	:param list only_show:
		Specifies which workspaces to show. Valid entries are ``"visible"``,
		``"urgent"`` and ``"focused"``. If omitted or ``null`` all workspaces
		are shown.
	:param str output:
		May be set to the name of an X output. If specified, only workspaces
		on that output are shown. Overrides automatic output detection by
		the lemonbar renderer and bindings.
		Use "__all__" to show workspaces on all outputs.
	:param int strip:
		Specifies how many characters from the front of each workspace name
		should be stripped (e.g. to remove workspace numbers). Defaults to zero.
	:param str format:
		Specifies the format used to display workspaces; defaults to ``{name}``.
		Valid fields are: ``name`` (workspace name), ``number`` (workspace number
		if present), `stipped_name`` (workspace name stripped of leading number),
		``icon`` (if available, icon for application running in the workspace,
		uses the ``multiple`` icon instead of multiple different icons), ``multi_icon``
		(similar to ``icon``, but does not use ``multiple``, instead joins all icons
		with a single space)
	:param dict icons:
		A dictionary mapping a substring of window classes to strings to be used as an
		icon for that window class.
		Further, there is a ``multiple`` icon for workspaces containing more than one
		window.
	:param bool sort_workspaces:
		Sort the workspaces displayed by their name according to the natural ordering.
	:param bool show_output:
		Shows the name of the output if more than one output is connected.
	:param list priority_workspaces:
		A list of workspace names to be sorted before any other workspaces in the given
		order.
	:param bool hide_empty_workspaces:
		Hides all workspaces without any open window.
		Also hides non-focussed workspaces containing only an open scratchpad.


	Highlight groups used: ``workspace`` or ``w_visible``, ``workspace:visible``, ``workspace`` or ``w_focused``, ``workspace:focused``, ``workspace`` or ``w_urgent``, ``workspace:urgent``, ``workspace`` or ``output``.
	'''
	conn = get_i3_connection()

	if not output == "__all__":
		output = output or segment_info.get('output')
	else:
		output = None

	if output:
		output = [output]
	else:
		output = [o.name for o in conn.get_outputs() if o.active]


	def sort_ws(ws):
		if sort_workspaces:
			def natural_key(ws):
				str = ws.name
				return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', str)]
			ws = sorted(ws, key=natural_key)
		result = []
		for n in priority_workspaces:
			result += [w for w in ws if w.name == n]
		return result + [w for w in ws if not w.name in priority_workspaces]

	ws_containers = {w_con.name : w_con for w_con in conn.get_tree().workspaces()}

	if len(output) <= 1:
		res = []
		if show_output:
			res += [{
				'contents': output[0],
				'highlight_groups': ['output']
			}]
		res += [{
			'contents': format.format(name = w.name[min(len(w.name), strip):],
				stripped_name = format_name(w.name, strip=True),
				number = w.num,
				icon = get_icon(w, '', icons, False, ws_containers),
				multi_icon = get_icon(w, ' ', icons, True, ws_containers)),
			'highlight_groups': workspace_groups(w)
			} for w in sort_ws(conn.get_workspaces()) \
					if (not only_show or any(getattr(w, tp) for tp in only_show)) \
					if w.output == output[0] \
					if not (hide_empty_workspaces and is_empty_workspace(w, ws_containers))]
		return res
	else:
		res = []
		for n in output:
			if show_output:
				res += [{
					'contents': n,
					'highlight_groups': ['output']
				}]
			res += [{
				'contents': format.format(name = w.name[min(len(w.name), strip):],
					stripped_name = format_name(w.name, strip=True),
					number = w.num,
					icon = get_icon(w, '', icons, False, ws_containers),
					multi_icon = get_icon(w, ' ', icons, True, ws_containers)),
				'highlight_groups': workspace_groups(w)
				} for w in sort_ws(conn.get_workspaces()) \
						if (not only_show or any(getattr(w, tp) for tp in only_show)) \
						if w.output == n \
						if not (hide_empty_workspaces and is_empty_workspace(w, ws_containers))]
		return res

@requires_segment_info
def workspace(pl, segment_info, workspace=None, strip=False, format=None, icons=WS_ICONS):
	'''Return the specified workspace name

	:param str workspace:
		Specifies which workspace to show. If unspecified, may be set by the
		``list_workspaces`` lister if used, otherwise falls back to
		currently focused workspace.

	:param bool strip:
		Specifies whether workspace numbers (in the ``1: name`` format) should
		be stripped from workspace names before being displayed. Defaults to false.
		Deprecated: Use {name} or {stripped_name} of format instead.

	:param str format:
		Specifies the format used to display workspaces; defaults to ``{name}``.
		Valid fields are: ``name`` (workspace name), ``number`` (workspace number
		if present), `stipped_name`` (workspace name stripped of leading number),
		``icon`` (if available, icon for application running in the workspace,
		uses the ``multiple`` icon instead of multiple different icons), ``multi_icon``
		(similar to ``icon``, but does not use ``multiple``, instead joins all icons
		with a single space)

	:param dict icons:
		A dictionary mapping a substring of window classes to strings to be used as an
		icon for that window class.
		Further, there is a ``multiple`` icon for workspaces containing more than one
		window.

	Highlight groups used: ``workspace`` or ``w_visible``, ``workspace:visible``, ``workspace`` or ``w_focused``, ``workspace:focused``, ``workspace`` or ``w_urgent``, ``workspace:urgent``, ``workspace``.
	'''
	if format == None:
		format = '{stripped_name}' if strip else '{name}'

	conn = get_i3_connection()
	ws_containers = {w_con.name : w_con for w_con in conn.get_tree().workspaces()}

	if workspace:
		try:
			w = next((
				w for w in conn.get_workspaces()
				if w.name == workspace
			))
		except StopIteration:
			return None
	elif segment_info.get('workspace'):
		w = segment_info['workspace']
	else:
		try:
			w = next((
				w for w in conn.get_workspaces()
				if w.focused
			))
		except StopIteration:
			return None

	return [{
		'contents': format.format(name = w.name,
			stripped_name = format_name(w.name, strip=True),
			number = w.num,
			icon = get_icon(w, '', icons, False, ws_containers),
			multi_icon = get_icon(w, ' ', icons, True, ws_containers)),
		'highlight_groups': workspace_groups(w)
		}]


@requires_segment_info
def mode(pl, segment_info, names={'default': None}):
	'''Returns current i3 mode

	:param dict names:
		Specifies the string to show for various modes.
		Use ``null`` to hide a mode (``default`` is hidden by default).

	Highligh groups used: ``mode``
	'''
	mode = segment_info['mode']
	if mode in names:
		return names[mode]
	return mode


def scratchpad_groups(w):
	group = []
	if w.urgent:
		group.append('scratchpad:urgent')
	if w.nodes[0].focused:
		group.append('scratchpad:focused')
	if w.workspace().name != '__i3_scratch':
		group.append('scratchpad:visible')
	group.append('scratchpad')
	return group


SCRATCHPAD_ICONS = {
	'fresh': 'O',
	'changed': 'X',
}


def scratchpad(pl, icons=SCRATCHPAD_ICONS):
	'''Returns the windows currently on the scratchpad

	:param dict icons:
		Specifies the strings to show for the different scratchpad window states. Must
		contain the keys ``fresh`` and ``changed``.

	Highlight groups used: ``scratchpad`` or ``scratchpad:visible``, ``scratchpad`` or ``scratchpad:focused``, ``scratchpad`` or ``scratchpad:urgent``.
	'''

	return [
		{
			'contents': icons.get(w.scratchpad_state, icons['changed']),
			'highlight_groups': scratchpad_groups(w)
		}
		for w in get_i3_connection().get_tree().descendants()
		if w.scratchpad_state != 'none'
	]

def active_window(pl, cutoff=100):
	'''Returns the title of the currently active window.

	:param int cutoff:
		Maximum title length. If the title is longer, the window_class is used instead.

	Highlight groups used: ``active_window_title``.
	'''

	focused = get_i3_connection().get_tree().find_focused()
	cont = focused.name
	if len(cont) > cutoff:
		cont = focused.window_class

	return [{'contents': cont, 'highlight_groups': ['active_window_title']}] if focused.name != focused.workspace().name else []



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import os
import errno
import ctypes
import struct

from ctypes.util import find_library

from powerline.lib.encoding import get_preferred_file_name_encoding


__copyright__ = '2013, Kovid Goyal <kovid at kovidgoyal.net>'
__docformat__ = 'restructuredtext en'


class INotifyError(Exception):
	pass


_inotify = None


def load_inotify():
	''' Initialize the inotify library '''
	global _inotify
	if _inotify is None:
		if hasattr(sys, 'getwindowsversion'):
			# On windows abort before loading the C library. Windows has
			# multiple, incompatible C runtimes, and we have no way of knowing
			# if the one chosen by ctypes is compatible with the currently
			# loaded one.
			raise INotifyError('INotify not available on windows')
		if sys.platform == 'darwin':
			raise INotifyError('INotify not available on OS X')
		if not hasattr(ctypes, 'c_ssize_t'):
			raise INotifyError('You need python >= 2.7 to use inotify')
		name = find_library('c')
		if not name:
			raise INotifyError('Cannot find C library')
		libc = ctypes.CDLL(name, use_errno=True)
		for function in ('inotify_add_watch', 'inotify_init1', 'inotify_rm_watch'):
			if not hasattr(libc, function):
				raise INotifyError('libc is too old')
		# inotify_init1()
		prototype = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, use_errno=True)
		init1 = prototype(('inotify_init1', libc), ((1, 'flags', 0),))

		# inotify_add_watch()
		prototype = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.c_char_p, ctypes.c_uint32, use_errno=True)
		add_watch = prototype(('inotify_add_watch', libc), (
			(1, 'fd'), (1, 'pathname'), (1, 'mask')))

		# inotify_rm_watch()
		prototype = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.c_int, use_errno=True)
		rm_watch = prototype(('inotify_rm_watch', libc), (
			(1, 'fd'), (1, 'wd')))

		# read()
		prototype = ctypes.CFUNCTYPE(ctypes.c_ssize_t, ctypes.c_int, ctypes.c_void_p, ctypes.c_size_t, use_errno=True)
		read = prototype(('read', libc), (
			(1, 'fd'), (1, 'buf'), (1, 'count')))
		_inotify = (init1, add_watch, rm_watch, read)
	return _inotify


class INotify(object):

	# See <sys/inotify.h> for the flags defined below

	# Supported events suitable for MASK parameter of INOTIFY_ADD_WATCH.
	ACCESS = 0x00000001         # File was accessed.
	MODIFY = 0x00000002         # File was modified.
	ATTRIB = 0x00000004         # Metadata changed.
	CLOSE_WRITE = 0x00000008    # Writtable file was closed.
	CLOSE_NOWRITE = 0x00000010  # Unwrittable file closed.
	OPEN = 0x00000020           # File was opened.
	MOVED_FROM = 0x00000040     # File was moved from X.
	MOVED_TO = 0x00000080       # File was moved to Y.
	CREATE = 0x00000100         # Subfile was created.
	DELETE = 0x00000200         # Subfile was deleted.
	DELETE_SELF = 0x00000400    # Self was deleted.
	MOVE_SELF = 0x00000800      # Self was moved.

	# Events sent by the kernel.
	UNMOUNT = 0x00002000     # Backing fs was unmounted.
	Q_OVERFLOW = 0x00004000  # Event queued overflowed.
	IGNORED = 0x00008000     # File was ignored.

	# Helper events.
	CLOSE = (CLOSE_WRITE | CLOSE_NOWRITE)  # Close.
	MOVE = (MOVED_FROM | MOVED_TO)         # Moves.

	# Special flags.
	ONLYDIR = 0x01000000      # Only watch the path if it is a directory.
	DONT_FOLLOW = 0x02000000  # Do not follow a sym link.
	EXCL_UNLINK = 0x04000000  # Exclude events on unlinked objects.
	MASK_ADD = 0x20000000     # Add to the mask of an already existing watch.
	ISDIR = 0x40000000        # Event occurred against dir.
	ONESHOT = 0x80000000      # Only send event once.

	# All events which a program can wait on.
	ALL_EVENTS = (
		ACCESS | MODIFY | ATTRIB | CLOSE_WRITE | CLOSE_NOWRITE | OPEN |
		MOVED_FROM | MOVED_TO | CREATE | DELETE | DELETE_SELF | MOVE_SELF
	)

	# See <bits/inotify.h>
	CLOEXEC = 0x80000
	NONBLOCK = 0x800

	def __init__(self, cloexec=True, nonblock=True):
		self._init1, self._add_watch, self._rm_watch, self._read = load_inotify()
		flags = 0
		if cloexec:
			flags |= self.CLOEXEC
		if nonblock:
			flags |= self.NONBLOCK
		self._inotify_fd = self._init1(flags)
		if self._inotify_fd == -1:
			raise INotifyError(os.strerror(ctypes.get_errno()))

		self._buf = ctypes.create_string_buffer(5000)
		self.fenc = get_preferred_file_name_encoding()
		self.hdr = struct.Struct(b'iIII')
		# We keep a reference to os to prevent it from being deleted
		# during interpreter shutdown, which would lead to errors in the
		# __del__ method
		self.os = os

	def handle_error(self):
		eno = ctypes.get_errno()
		extra = ''
		if eno == errno.ENOSPC:
			extra = 'You may need to increase the inotify limits on your system, via /proc/sys/fs/inotify/max_user_*'
		raise OSError(eno, self.os.strerror(eno) + str(extra))

	def __del__(self):
		# This method can be called during interpreter shutdown, which means we
		# must do the absolute minimum here. Note that there could be running
		# daemon threads that are trying to call other methods on this object.
		try:
			self.os.close(self._inotify_fd)
		except (AttributeError, TypeError):
			pass

	def close(self):
		if hasattr(self, '_inotify_fd'):
			self.os.close(self._inotify_fd)
			del self.os
			del self._add_watch
			del self._rm_watch
			del self._inotify_fd

	def read(self, get_name=True):
		buf = []
		while True:
			num = self._read(self._inotify_fd, self._buf, len(self._buf))
			if num == 0:
				break
			if num < 0:
				en = ctypes.get_errno()
				if en == errno.EAGAIN:
					break  # No more data
				if en == errno.EINTR:
					continue  # Interrupted, try again
				raise OSError(en, self.os.strerror(en))
			buf.append(self._buf.raw[:num])
		raw = b''.join(buf)
		pos = 0
		lraw = len(raw)
		while lraw - pos >= self.hdr.size:
			wd, mask, cookie, name_len = self.hdr.unpack_from(raw, pos)
			pos += self.hdr.size
			name = None
			if get_name:
				name = raw[pos:pos + name_len].rstrip(b'\0')
			pos += name_len
			self.process_event(wd, mask, cookie, name)

	def process_event(self, *args):
		raise NotImplementedError()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)


class Node(object):
	def __init__(self, tag, value, start_mark, end_mark):
		self.tag = tag
		self.value = value
		self.start_mark = start_mark
		self.end_mark = end_mark

	def __repr__(self):
		value = self.value
		# if isinstance(value, list):
		# 	if len(value) == 0:
		# 		value = '<empty>'
		# 	elif len(value) == 1:
		# 		value = '<1 item>'
		# 	else:
		# 		value = '<%d items>' % len(value)
		# else:
		# 	if len(value) > 75:
		# 		value = repr(value[:70]+u' ... ')
		# 	else:
		# 		value = repr(value)
		value = repr(value)
		return '%s(tag=%r, value=%s)' % (self.__class__.__name__, self.tag, value)


class ScalarNode(Node):
	id = 'scalar'

	def __init__(self, tag, value, start_mark=None, end_mark=None, style=None):
		self.tag = tag
		self.value = value
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.style = style


class CollectionNode(Node):
	def __init__(self, tag, value, start_mark=None, end_mark=None, flow_style=None):
		self.tag = tag
		self.value = value
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.flow_style = flow_style


class SequenceNode(CollectionNode):
	id = 'sequence'


class MappingNode(CollectionNode):
	id = 'mapping'



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)


class Token(object):
	def __init__(self, start_mark, end_mark):
		self.start_mark = start_mark
		self.end_mark = end_mark

	def __repr__(self):
		attributes = [
			key for key in self.__dict__
			if not key.endswith('_mark')
		]
		attributes.sort()
		arguments = ', '.join([
			'%s=%r' % (key, getattr(self, key))
			for key in attributes
		])
		return '%s(%s)' % (self.__class__.__name__, arguments)


class StreamStartToken(Token):
	id = '<stream start>'

	def __init__(self, start_mark=None, end_mark=None, encoding=None):
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.encoding = encoding


class StreamEndToken(Token):
	id = '<stream end>'


class FlowSequenceStartToken(Token):
	id = '['


class FlowMappingStartToken(Token):
	id = '{'


class FlowSequenceEndToken(Token):
	id = ']'


class FlowMappingEndToken(Token):
	id = '}'


class KeyToken(Token):
	id = '?'


class ValueToken(Token):
	id = ':'


class FlowEntryToken(Token):
	id = ','


class ScalarToken(Token):
	id = '<scalar>'

	def __init__(self, value, plain, start_mark, end_mark, style=None):
		self.value = value
		self.plain = plain
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.style = style



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os

from threading import RLock

from powerline.lib.path import realpath


class StatFileWatcher(object):
	def __init__(self):
		self.watches = {}
		self.lock = RLock()

	def watch(self, path):
		path = realpath(path)
		with self.lock:
			self.watches[path] = os.path.getmtime(path)

	def unwatch(self, path):
		path = realpath(path)
		with self.lock:
			self.watches.pop(path, None)

	def is_watching(self, path):
		with self.lock:
			return realpath(path) in self.watches

	def __call__(self, path):
		path = realpath(path)
		with self.lock:
			if path not in self.watches:
				self.watches[path] = os.path.getmtime(path)
				return True
			mtime = os.path.getmtime(path)
			if mtime != self.watches[path]:
				self.watches[path] = mtime
				return True
			return False

	def close(self):
		with self.lock:
			self.watches.clear()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderers.shell.zsh import ZshPromptRenderer


class TcshPromptRenderer(ZshPromptRenderer):
	'''Powerline tcsh prompt segment renderer.'''
	character_translations = ZshPromptRenderer.character_translations.copy()
	character_translations[ord('%')] = '%%'
	character_translations[ord('\\')] = '\\\\'
	character_translations[ord('^')] = '\\^'
	character_translations[ord('!')] = '\\!'

	def do_render(self, **kwargs):
		ret = super(TcshPromptRenderer, self).do_render(**kwargs)
		nbsp = self.character_translations.get(ord(' '), ' ')
		end = self.hlstyle()
		assert not ret or ret.endswith(end)
		if ret.endswith(nbsp + end):
			# Exchange nbsp and highlight end because tcsh removes trailing 
			# %{%} part of the prompt for whatever reason
			ret = ret[:-(len(nbsp) + len(end))] + end + nbsp
		else:
			# We *must* end prompt with non-%{%} sequence for the reasons 
			# explained above. So add nbsp if it is not already there.
			ret += nbsp
		return ret


renderer = TcshPromptRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from weakref import ref

from IPython.terminal.prompts import Prompts
from pygments.token import Token  # NOQA

from powerline.ipython import IPythonPowerline
from powerline.renderers.ipython.since_5 import PowerlinePromptStyle
from powerline.bindings.ipython.post_0_11 import PowerlineMagics, ShutdownHook


class ConfigurableIPythonPowerline(IPythonPowerline):
	def init(self, ip):
		config = ip.config.Powerline
		self.config_overrides = config.get('config_overrides')
		self.theme_overrides = config.get('theme_overrides', {})
		self.config_paths = config.get('config_paths')
		super(ConfigurableIPythonPowerline, self).init(
			renderer_module='.since_5')

	def do_setup(self, ip, prompts, shutdown_hook):
		prompts.powerline = self

		msfn_missing = ()
		saved_msfn = getattr(ip, '_make_style_from_name', msfn_missing)

		if hasattr(saved_msfn, 'powerline_original'):
			saved_msfn = saved_msfn.powerline_original

		def _make_style_from_name(ip, name):
			prev_style = saved_msfn(name)
			new_style = PowerlinePromptStyle(lambda: prev_style)
			return new_style

		_make_style_from_name.powerline_original = saved_msfn

		if not isinstance(ip._style, PowerlinePromptStyle):
			prev_style = ip._style
			ip._style = PowerlinePromptStyle(lambda: prev_style)

		if not isinstance(saved_msfn, type(self.init)):
			_saved_msfn = saved_msfn
			saved_msfn = lambda: _saved_msfn(ip)

		if saved_msfn is not msfn_missing:
			ip._make_style_from_name = _make_style_from_name

		magics = PowerlineMagics(ip, self)
		ip.register_magics(magics)

		if shutdown_hook:
			shutdown_hook.powerline = ref(self)


class PowerlinePrompts(Prompts):
	'''Class that returns powerline prompts
	'''
	def __init__(self, shell):
		shutdown_hook = ShutdownHook(shell)
		powerline = ConfigurableIPythonPowerline(shell)
		self.shell = shell
		powerline.do_setup(shell, self, shutdown_hook)
		self.last_output_count = None
		self.last_output = {}

	for prompt in ('in', 'continuation', 'rewrite', 'out'):
		exec((
			'def {0}_prompt_tokens(self, *args, **kwargs):\n'
			'	if self.last_output_count != self.shell.execution_count:\n'
			'		self.last_output.clear()\n'
			'		self.last_output_count = self.shell.execution_count\n'
			'	if "{0}" not in self.last_output:\n'
			'		self.last_output["{0}"] = self.powerline.render('
			'			side="left",'
			'			matcher_info="{1}",'
			'			segment_info=self.shell,'
			'		) + [(Token.Generic.Prompt, " ")]\n'
			'	return self.last_output["{0}"]'
		).format(prompt, 'in2' if prompt == 'continuation' else prompt))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import itertools

from powerline.segment import gen_segment_getter, process_segment, get_fallback_segment
from powerline.lib.unicode import u, safe_unicode


def requires_segment_info(func):
	func.powerline_requires_segment_info = True
	return func


def requires_filesystem_watcher(func):
	func.powerline_requires_filesystem_watcher = True
	return func


def new_empty_segment_line():
	return {
		'left': [],
		'right': []
	}


def add_spaces_left(pl, amount, segment):
	return (' ' * amount) + segment['contents']


def add_spaces_right(pl, amount, segment):
	return segment['contents'] + (' ' * amount)


def add_spaces_center(pl, amount, segment):
	amount, remainder = divmod(amount, 2)
	return (' ' * (amount + remainder)) + segment['contents'] + (' ' * amount)


expand_functions = {
	'l': add_spaces_right,
	'r': add_spaces_left,
	'c': add_spaces_center,
}


class Theme(object):
	def __init__(self,
	             ext,
	             theme_config,
	             common_config,
	             pl,
	             get_module_attr,
	             top_theme,
	             colorscheme,
	             main_theme_config=None,
	             run_once=False,
	             shutdown_event=None):
		self.colorscheme = colorscheme
		self.dividers = theme_config['dividers']
		self.dividers = dict((
			(key, dict((k, u(v))
			for k, v in val.items()))
			for key, val in self.dividers.items()
		))
		try:
			self.cursor_space_multiplier = 1 - (theme_config['cursor_space'] / 100)
		except KeyError:
			self.cursor_space_multiplier = None
		self.cursor_columns = theme_config.get('cursor_columns')
		self.spaces = theme_config['spaces']
		self.outer_padding = int(theme_config.get('outer_padding', 1))
		self.segments = []
		self.EMPTY_SEGMENT = {
			'contents': None,
			'highlight': {'fg': False, 'bg': False, 'attrs': 0}
		}
		self.pl = pl
		theme_configs = [theme_config]
		if main_theme_config:
			theme_configs.append(main_theme_config)
		get_segment = gen_segment_getter(
			pl,
			ext,
			common_config,
			theme_configs,
			theme_config.get('default_module'),
			get_module_attr,
			top_theme
		)
		for segdict in itertools.chain((theme_config['segments'],),
		                               theme_config['segments'].get('above', ())):
			self.segments.append(new_empty_segment_line())
			for side in ['left', 'right']:
				for segment in segdict.get(side, []):
					segment = get_segment(segment, side)
					if segment:
						if not run_once:
							if segment['startup']:
								try:
									segment['startup'](pl, shutdown_event)
								except Exception as e:
									pl.error('Exception during {0} startup: {1}', segment['name'], str(e))
									continue
						self.segments[-1][side].append(segment)

	def shutdown(self):
		for line in self.segments:
			for segments in line.values():
				for segment in segments:
					try:
						segment['shutdown']()
					except TypeError:
						pass

	def get_divider(self, side='left', type='soft'):
		'''Return segment divider.'''
		return self.dividers[side][type]

	def get_spaces(self):
		return self.spaces

	def get_line_number(self):
		return len(self.segments)

	def get_segments(self, side=None, line=0, segment_info=None, mode=None):
		'''Return all segments.

		Function segments are called, and all segments get their before/after
		and ljust/rjust properties applied.

		:param int line:
			Line number for which segments should be obtained. Is counted from 
			zero (botmost line).
		'''
		for side in [side] if side else ['left', 'right']:
			parsed_segments = []
			for segment in self.segments[line][side]:
				if segment['display_condition'](self.pl, segment_info, mode):
					process_segment(
						self.pl,
						side,
						segment_info,
						parsed_segments,
						segment,
						mode,
						self.colorscheme,
					)
			for segment in parsed_segments:
				self.pl.prefix = segment['name']
				try:
					width = segment['width']
					align = segment['align']
					if width == 'auto' and segment['expand'] is None:
						segment['expand'] = expand_functions.get(align)
						if segment['expand'] is None:
							self.pl.error('Align argument must be “r”, “l” or “c”, not “{0}”', align)

					try:
						segment['contents'] = segment['before'] + u(
							segment['contents'] if segment['contents'] is not None else ''
						) + segment['after']
					except Exception as e:
						self.pl.exception('Failed to compute segment contents: {0}', str(e))
						segment['contents'] = safe_unicode(segment.get('contents'))
					# Align segment contents
					if segment['width'] and segment['width'] != 'auto':
						if segment['align'] == 'l':
							segment['contents'] = segment['contents'].ljust(segment['width'])
						elif segment['align'] == 'r':
							segment['contents'] = segment['contents'].rjust(segment['width'])
						elif segment['align'] == 'c':
							segment['contents'] = segment['contents'].center(segment['width'])
					# We need to yield a copy of the segment, or else mode-dependent
					# segment contents can’t be cached correctly e.g. when caching
					# non-current window contents for vim statuslines
					yield segment.copy()
				except Exception as e:
					self.pl.exception('Failed to compute segment: {0}', str(e))
					fallback = get_fallback_segment()
					fallback.update(side=side)
					yield fallback



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import operator

from collections import defaultdict

try:
	from __builtin__ import reduce
except ImportError:
	from functools import reduce

from pygments.token import Token
from prompt_toolkit.styles import DynamicStyle, Attrs

from powerline.renderers.ipython import IPythonRenderer
from powerline.ipython import IPythonInfo
from powerline.colorscheme import ATTR_BOLD, ATTR_ITALIC, ATTR_UNDERLINE


PowerlinePromptToken = Token.Generic.Prompt.Powerline


# Note: since 2.7 there is dict.__missing__ with same purpose. But in 2.6 one 
# must use defaultdict to get __missing__ working.
class PowerlineStyleDict(defaultdict):
	'''Dictionary used for getting pygments style for Powerline groups
	'''
	def __new__(cls, missing_func):
		return defaultdict.__new__(cls)

	def __init__(self, missing_func):
		super(PowerlineStyleDict, self).__init__()
		self.missing_func = missing_func

	def __missing__(self, key):
		return self.missing_func(key)


class PowerlinePromptStyle(DynamicStyle):
	def get_attrs_for_token(self, token):
		if (
			token not in PowerlinePromptToken
			or len(token) != len(PowerlinePromptToken) + 1
			or not token[-1].startswith('Pl')
			or token[-1] == 'Pl'
		):
			return super(PowerlinePromptStyle, self).get_attrs_for_token(token)
		ret = {
			'color': None,
			'bgcolor': None,
			'bold': None,
			'underline': None,
			'italic': None,
			'reverse': False,
			'blink': False,
		}
		for prop in token[-1][3:].split('_'):
			if prop[0] == 'a':
				ret[prop[1:]] = True
			elif prop[0] == 'f':
				ret['color'] = prop[1:]
			elif prop[0] == 'b':
				ret['bgcolor'] = prop[1:]
		return Attrs(**ret)

	def get_token_to_attributes_dict(self):
		dct = super(PowerlinePromptStyle, self).get_token_to_attributes_dict()

		def fallback(key):
			try:
				return dct[key]
			except KeyError:
				return self.get_attrs_for_token(key)

		return PowerlineStyleDict(fallback)

	def invalidation_hash(self):
		return super(PowerlinePromptStyle, self).invalidation_hash() + 1


class IPythonPygmentsRenderer(IPythonRenderer):
	reduce_initial = []

	def get_segment_info(self, segment_info, mode):
		return super(IPythonPygmentsRenderer, self).get_segment_info(
			IPythonInfo(segment_info), mode)

	@staticmethod
	def hl_join(segments):
		return reduce(operator.iadd, segments, [])

	def hl(self, contents, fg=None, bg=None, attrs=None, **kwargs):
		'''Output highlighted chunk.

		This implementation outputs a list containing a single pair 
		(:py:class:`pygments.token.Token`, 
		:py:class:`powerline.lib.unicode.unicode`).
		'''
		guifg = None
		guibg = None
		attrs = []
		if fg is not None and fg is not False:
			guifg = fg[1]
		if bg is not None and bg is not False:
			guibg = bg[1]
		if attrs:
			attrs = []
			if attrs & ATTR_BOLD:
				attrs.append('bold')
			if attrs & ATTR_ITALIC:
				attrs.append('italic')
			if attrs & ATTR_UNDERLINE:
				attrs.append('underline')
		name = (
			'Pl'
			+ ''.join(('_a' + attr for attr in attrs))
			+ (('_f%6x' % guifg) if guifg is not None else '')
			+ (('_b%6x' % guibg) if guibg is not None else '')
		)
		return [(getattr(Token.Generic.Prompt.Powerline, name), contents)]

	def hlstyle(self, **kwargs):
		return []

	def get_client_id(self, segment_info):
		return id(self)


renderer = IPythonPygmentsRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.lib.vcs import guess, tree_status
from powerline.segments import Segment, with_docstring
from powerline.theme import requires_segment_info, requires_filesystem_watcher


@requires_filesystem_watcher
@requires_segment_info
class BranchSegment(Segment):
	divider_highlight_group = None

	@staticmethod
	def get_directory(segment_info):
		return segment_info['getcwd']()

	def __call__(self, pl, segment_info, create_watcher, status_colors=False, ignore_statuses=()):
		name = self.get_directory(segment_info)
		if name:
			repo = guess(path=name, create_watcher=create_watcher)
			if repo is not None:
				branch = repo.branch()
				scol = ['branch']
				if status_colors:
					try:
						status = tree_status(repo, pl)
					except Exception as e:
						pl.exception('Failed to compute tree status: {0}', str(e))
						status = '?'
					else:
						status = status and status.strip()
						if status in ignore_statuses:
							status = None
					scol.insert(0, 'branch_dirty' if status else 'branch_clean')
				return [{
					'contents': branch,
					'highlight_groups': scol,
					'divider_highlight_group': self.divider_highlight_group,
				}]


branch = with_docstring(BranchSegment(),
'''Return the current VCS branch.

:param bool status_colors:
	Determines whether repository status will be used to determine highlighting. 
	Default: False.
:param list ignore_statuses:
	List of statuses which will not result in repo being marked as dirty. Most 
	useful is setting this option to ``["U"]``: this will ignore repository 
	which has just untracked files (i.e. repository with modified, deleted or 
	removed files will be marked as dirty, while just untracked files will make 
	segment show clean repository). Only applicable if ``status_colors`` option 
	is True.

Highlight groups used: ``branch_clean``, ``branch_dirty``, ``branch``.
''')


@requires_filesystem_watcher
@requires_segment_info
class StashSegment(Segment):
	divider_highlight_group = None

	@staticmethod
	def get_directory(segment_info):
		return segment_info['getcwd']()

	def __call__(self, pl, segment_info, create_watcher):
		name = self.get_directory(segment_info)
		if name:
			repo = guess(path=name, create_watcher=create_watcher)
			if repo is not None:
				stash = getattr(repo, 'stash', None)
				if stash:
					stashes = stash()
					if stashes:
						return [{
							'contents': str(stashes),
							'highlight_groups': ['stash'],
							'divider_highlight_group': self.divider_highlight_group
						}]

stash = with_docstring(StashSegment(),
'''Return the number of current VCS stash entries, if any.

Highlight groups used: ``stash``.
''')



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from functools import wraps

from powerline.lib.monotonic import monotonic


def default_cache_key(**kwargs):
	return frozenset(kwargs.items())


class memoize(object):
	'''Memoization decorator with timeout.'''
	def __init__(self, timeout, cache_key=default_cache_key, cache_reg_func=None):
		self.timeout = timeout
		self.cache_key = cache_key
		self.cache = {}
		self.cache_reg_func = cache_reg_func

	def __call__(self, func):
		@wraps(func)
		def decorated_function(**kwargs):
			if self.cache_reg_func:
				self.cache_reg_func(self.cache)
				self.cache_reg_func = None

			key = self.cache_key(**kwargs)
			try:
				cached = self.cache.get(key, None)
			except TypeError:
				return func(**kwargs)
			# Handle case when time() appears to be less then cached['time'] due 
			# to clock updates. Not applicable for monotonic clock, but this 
			# case is currently rare.
			if cached is None or not (cached['time'] < monotonic() < cached['time'] + self.timeout):
				cached = self.cache[key] = {
					'result': func(**kwargs),
					'time': monotonic(),
				}
			return cached['result']
		return decorated_function



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import gc
import sys

from types import FrameType
from itertools import chain


# From http://code.activestate.com/recipes/523004-find-cyclical-references/
def print_cycles(objects, outstream=sys.stdout, show_progress=False):
	'''Find reference cycles

	:param list objects:
		A list of objects to find cycles in.  It is often useful to pass in 
		gc.garbage to find the cycles that are preventing some objects from 
		being garbage collected.
	:param file outstream:
		The stream for output.
	:param bool show_progress:
		If True, print the number of objects reached as they are found.
	'''
	def print_path(path):
		for i, step in enumerate(path):
			# next “wraps around”
			next = path[(i + 1) % len(path)]

			outstream.write('	%s -- ' % str(type(step)))
			written = False
			if isinstance(step, dict):
				for key, val in step.items():
					if val is next:
						outstream.write('[%s]' % repr(key))
						written = True
						break
					if key is next:
						outstream.write('[key] = %s' % repr(val))
						written = True
						break
			elif isinstance(step, (list, tuple)):
				for i, item in enumerate(step):
					if item is next:
						outstream.write('[%d]' % i)
						written = True
			elif getattr(type(step), '__getattribute__', None) in (object.__getattribute__, type.__getattribute__):
				for attr in chain(dir(step), getattr(step, '__dict__', ())):
					if getattr(step, attr, None) is next:
						try:
							outstream.write('%r.%s' % (step, attr))
						except TypeError:
							outstream.write('.%s' % (step, attr))
						written = True
						break
			if not written:
				outstream.write(repr(step))
			outstream.write(' ->\n')
		outstream.write('\n')

	def recurse(obj, start, all, current_path):
		if show_progress:
			outstream.write('%d\r' % len(all))

		all[id(obj)] = None

		referents = gc.get_referents(obj)
		for referent in referents:
			# If we’ve found our way back to the start, this is
			# a cycle, so print it out
			if referent is start:
				try:
					outstream.write('Cyclic reference: %r\n' % referent)
				except TypeError:
					try:
						outstream.write('Cyclic reference: %i (%r)\n' % (id(referent), type(referent)))
					except TypeError:
						outstream.write('Cyclic reference: %i\n' % id(referent))
				print_path(current_path)

			# Don’t go back through the original list of objects, or
			# through temporary references to the object, since those
			# are just an artifact of the cycle detector itself.
			elif referent is objects or isinstance(referent, FrameType): 
				continue

			# We haven’t seen this object before, so recurse
			elif id(referent) not in all:
				recurse(referent, start, all, current_path + (obj,))

	for obj in objects:
		# We are not interested in non-powerline cyclic references
		try:
			if not type(obj).__module__.startswith('powerline'):
				continue
		except AttributeError:
			continue
		recurse(obj, obj, {}, ())



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.lint.markedjson.reader import Reader
from powerline.lint.markedjson.scanner import Scanner
from powerline.lint.markedjson.parser import Parser
from powerline.lint.markedjson.composer import Composer
from powerline.lint.markedjson.constructor import Constructor
from powerline.lint.markedjson.resolver import Resolver
from powerline.lint.markedjson.error import echoerr


class Loader(Reader, Scanner, Parser, Composer, Constructor, Resolver):
	def __init__(self, stream):
		Reader.__init__(self, stream)
		Scanner.__init__(self)
		Parser.__init__(self)
		Composer.__init__(self)
		Constructor.__init__(self)
		Resolver.__init__(self)
		self.haserrors = False

	def echoerr(self, *args, **kwargs):
		echoerr(*args, **kwargs)
		self.haserrors = True



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import re

from powerline.theme import requires_segment_info
from powerline.lib.shell import run_cmd
from powerline.bindings.wm.awesome import AwesomeThread


DEFAULT_UPDATE_INTERVAL = 0.5


conn = None


def i3_subscribe(conn, event, callback):
	'''Subscribe to i3 workspace event

	:param conn:
		Connection returned by :py:func:`get_i3_connection`.
	:param str event:
		Event to subscribe to, e.g. ``'workspace'``.
	:param func callback:
		Function to run on event.
	'''
	conn.on(event, callback)

	from threading import Thread

	class I3Thread(Thread):
		daemon = True

		def __init__(self, conn):
			super(I3Thread, self).__init__()
			self.__conn = conn

		def run(self):
			self.__conn.main()

	thread = I3Thread(conn=conn)

	thread.start()


def get_i3_connection():
	'''Return a valid, cached i3 Connection instance
	'''
	global conn
	if not conn:
		import i3ipc
		conn = i3ipc.Connection()
	return conn


XRANDR_OUTPUT_RE = re.compile(r'^(?P<name>[0-9A-Za-z-]+) connected(?P<primary> primary)? (?P<width>\d+)x(?P<height>\d+)\+(?P<x>\d+)\+(?P<y>\d+)', re.MULTILINE)


def get_connected_xrandr_outputs(pl):
	'''Iterate over xrandr outputs

	Outputs are represented by a dictionary with ``name``, ``width``, 
	``height``, ``primary``, ``x`` and ``y`` keys.
	'''
	return (match.groupdict() for match in XRANDR_OUTPUT_RE.finditer(
	    run_cmd(pl, ['xrandr', '-q'])
	))


wm_threads = {
	'awesome': AwesomeThread,
}



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division,
                        absolute_import, print_function)

from weakref import ref
from warnings import warn

try:
	from IPython.core.prompts import PromptManager
	has_prompt_manager = True
except ImportError:
	has_prompt_manager = False
from IPython.core.magic import Magics, magics_class, line_magic

from powerline.ipython import IPythonPowerline, IPythonInfo

if has_prompt_manager:
	from powerline.ipython import RewriteResult


@magics_class
class PowerlineMagics(Magics):
	def __init__(self, ip, powerline):
		super(PowerlineMagics, self).__init__(ip)
		self._powerline = powerline

	@line_magic
	def powerline(self, line):
		if line == 'reload':
			self._powerline.reload()
		else:
			raise ValueError('Expected `reload`, but got {0}'.format(line))


old_prompt_manager = None


class ShutdownHook(object):
	def __init__(self, ip):
		self.powerline = lambda: None
		ip.hooks.shutdown_hook.add(self)

	def __call__(self):
		from IPython.core.hooks import TryNext
		powerline = self.powerline()
		if powerline is not None:
			powerline.shutdown()
		raise TryNext()


if has_prompt_manager:
	class PowerlinePromptManager(PromptManager):
		def __init__(self, powerline, shell):
			self.powerline = powerline
			self.powerline_segment_info = IPythonInfo(shell)
			self.shell = shell

		def render(self, name, color=True, *args, **kwargs):
			res = self.powerline.render(
				is_prompt=name.startswith('in'),
				side='left',
				output_width=True,
				output_raw=not color,
				matcher_info=name,
				segment_info=self.powerline_segment_info,
			)
			self.txtwidth = res[-1]
			self.width = res[-1]
			ret = res[0] if color else res[1]
			if name == 'rewrite':
				return RewriteResult(ret)
			else:
				return ret

	class ConfigurableIPythonPowerline(IPythonPowerline):
		def init(self, ip):
			config = ip.config.Powerline
			self.config_overrides = config.get('config_overrides')
			self.theme_overrides = config.get('theme_overrides', {})
			self.config_paths = config.get('config_paths')
			if has_prompt_manager:
				renderer_module = '.pre_5'
			else:
				renderer_module = '.since_7'
			super(ConfigurableIPythonPowerline, self).init(
				renderer_module=renderer_module)

		def do_setup(self, ip, shutdown_hook):
			global old_prompt_manager

			if old_prompt_manager is None:
				old_prompt_manager = ip.prompt_manager
			prompt_manager = PowerlinePromptManager(
				powerline=self,
				shell=ip.prompt_manager.shell,
			)
			ip.prompt_manager = prompt_manager

			magics = PowerlineMagics(ip, self)
			shutdown_hook.powerline = ref(self)
			ip.register_magics(magics)


def load_ipython_extension(ip):
	if has_prompt_manager:
		shutdown_hook = ShutdownHook(ip)
		powerline = ConfigurableIPythonPowerline(ip)
		powerline.setup(ip, shutdown_hook)
	else:
		from powerline.bindings.ipython.since_7 import PowerlinePrompts
		ip.prompts_class = PowerlinePrompts
		ip.prompts = PowerlinePrompts(ip)
		warn(DeprecationWarning(
			'post_0_11 extension is deprecated since IPython 5, use\n'
			'  from powerline.bindings.ipython.since_7 import PowerlinePrompts\n'
			'  c.TerminalInteractiveShell.prompts_class = PowerlinePrompts\n'
			'or check: \n'
			'https://powerline.readthedocs.io/en/master/usage/other.html\n'
		))


def unload_ipython_extension(ip):
	global old_prompt_manager
	if old_prompt_manager is not None:
		ip.prompt_manager = old_prompt_manager
	old_prompt_manager = None



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

try:
	import vim
except ImportError:
	vim = object()

from powerline.bindings.vim import vim_global_exists
from powerline.theme import requires_segment_info


@requires_segment_info
def ale(segment_info, pl, err_format='ERR: ln {first_line} ({num}) ', warn_format='WARN: ln {first_line} ({num}) '):
	'''Show whether ALE has found any errors or warnings

	:param str err_format:
		Format string for errors.

	:param str warn_format:
		Format string for warnings.

	Highlight groups used: ``ale:warning`` or ``warning``, ``ale:error`` or ``error``.
	'''
	if not (vim_global_exists('ale_enabled') and int(vim.eval('g:ale_enabled'))):
		return None
	has_errors = int(vim.eval('ale#statusline#Count(' + str(segment_info['bufnr']) + ').total'))
	if not has_errors:
		return
	error = None
	warning = None
	errors_count = 0
	warnings_count = 0
	for issue in vim.eval('ale#engine#GetLoclist(' + str(segment_info['bufnr']) + ')'):
		if issue['type'] == 'E':
			error = error or issue
			errors_count += 1
		elif issue['type'] == 'W':
			warning = warning or issue
			warnings_count += 1
	segments = []
	if error:
		segments.append({
			'contents': err_format.format(first_line=error['lnum'], num=errors_count),
			'highlight_groups': ['ale:error', 'error'],
		})
	if warning:
		segments.append({
			'contents': warn_format.format(first_line=warning['lnum'], num=warnings_count),
			'highlight_groups': ['ale:warning', 'warning'],
		})
	return segments



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.theme import requires_segment_info
from powerline.segments import with_docstring
from powerline.segments.common.env import CwdSegment
from powerline.lib.unicode import out_u


@requires_segment_info
def jobnum(pl, segment_info, show_zero=False):
	'''Return the number of jobs.

	:param bool show_zero:
		If False (default) shows nothing if there are no jobs. Otherwise shows
		zero for no jobs.
	'''
	jobnum = segment_info['args'].jobnum
	if jobnum is None or (not show_zero and jobnum == 0):
		return None
	else:
		return str(jobnum)

try:
	import signal
	exit_codes = dict((k, v) for v, k in reversed(sorted(signal.__dict__.items())) \
		if v.startswith('SIG') and not v.startswith('SIG_'))
except ImportError:
	exit_codes = dict()

@requires_segment_info
def last_status(pl, segment_info, signal_names=True):
	'''Return last exit code.

	:param bool signal_names:
		If True (default), translate signal numbers to human-readable names.

	Highlight groups used: ``exit_fail``
	'''
	if not segment_info['args'].last_exit_code:
		return None

	try:
		if signal_names and segment_info['args'].last_exit_code - 128 in exit_codes:
			return [{'contents': exit_codes[segment_info['args'].last_exit_code - 128], 'highlight_groups': ['exit_fail']}]
	except TypeError:
		pass
	return [{'contents': str(segment_info['args'].last_exit_code), 'highlight_groups': ['exit_fail']}]


@requires_segment_info
def last_pipe_status(pl, segment_info, signal_names=True):
	'''Return last pipe status.

	:param bool signal_names:
		If True (default), translate signal numbers to human-readable names.

	Highlight groups used: ``exit_fail``, ``exit_success``
	'''
	last_pipe_status = (
		segment_info['args'].last_pipe_status
		or (segment_info['args'].last_exit_code,)
	)
	if any(last_pipe_status):
		try:
			return [{
				'contents': exit_codes[status - 128] if signal_names and \
					status - 128 in exit_codes else str(status),
				'highlight_groups': ['exit_fail' if status else 'exit_success'],
				'draw_inner_divider': True
			} for status in last_pipe_status]
		except TypeError:
			return [{
				'contents': str(status),
				'highlight_groups': ['exit_fail' if status else 'exit_success'],
				'draw_inner_divider': True
			} for status in last_pipe_status]
	else:
		return None

@requires_segment_info
def mode(pl, segment_info, override={'vicmd': 'COMMND', 'viins': 'INSERT'}, default=None):
	'''Return the current mode.

	:param dict override:
		dict for overriding mode strings.
	:param str default:
		If current mode is equal to this string then this segment will not get
		displayed. If not specified the value is taken from
		``$POWERLINE_DEFAULT_MODE`` variable. This variable is set by zsh
		bindings for any mode that does not start from ``vi``.
	'''
	mode = segment_info.get('mode', None)
	if not mode:
		pl.debug('No mode specified')
		return None
	default = default or segment_info.get('default_mode', None)
	if mode == default:
		return None
	try:
		return override[mode]
	except KeyError:
		# Note: with zsh line editor you can emulate as much modes as you wish.
		# Thus having unknown mode is not an error: maybe just some developer
		# added support for his own zle widgets. As there is no built-in mode()
		# function like in VimL and mode is likely be defined by our code or by
		# somebody knowing what he is doing there is absolutely no need in
		# keeping translations dictionary.
		return mode.upper()


@requires_segment_info
def continuation(pl, segment_info, omit_cmdsubst=True, right_align=False, renames={}):
	'''Display parser state.

	:param bool omit_cmdsubst:
		Do not display cmdsubst parser state if it is the last one.
	:param bool right_align:
		Align to the right.
	:param dict renames:
		Rename states: ``{old_name : new_name}``. If ``new_name`` is ``None``
		then given state is not displayed.

	Highlight groups used: ``continuation``, ``continuation:current``.
	'''
	if not segment_info.get('parser_state'):
		return [{
			'contents': '',
			'width': 'auto',
			'highlight_groups': ['continuation:current', 'continuation'],
		}]
	ret = []

	for state in segment_info['parser_state'].split():
		state = renames.get(state, state)
		if state:
			ret.append({
				'contents': state,
				'highlight_groups': ['continuation'],
				'draw_inner_divider': True,
			})

	if omit_cmdsubst and ret[-1]['contents'] == 'cmdsubst':
		ret.pop(-1)

	if not ret:
		ret.append({
			'contents': ''
		})

	if right_align:
		ret[0].update(width='auto', align='r')
		ret[-1]['highlight_groups'] = ['continuation:current', 'continuation']
	else:
		ret[-1].update(width='auto', align='l', highlight_groups=['continuation:current', 'continuation'])

	return ret


@requires_segment_info
class ShellCwdSegment(CwdSegment):
	def get_shortened_path(self, pl, segment_info, use_shortened_path=True, **kwargs):
		if use_shortened_path:
			try:
				return out_u(segment_info['shortened_path'])
			except KeyError:
				pass
		return super(ShellCwdSegment, self).get_shortened_path(pl, segment_info, **kwargs)


cwd = with_docstring(ShellCwdSegment(),
'''Return the current working directory.

Returns a segment list to create a breadcrumb-like effect.

:param int dir_shorten_len:
	shorten parent directory names to this length (e.g.
	:file:`/long/path/to/powerline` → :file:`/l/p/t/powerline`)
:param int dir_limit_depth:
	limit directory depth to this number (e.g.
	:file:`/long/path/to/powerline` → :file:`⋯/to/powerline`)
:param bool use_path_separator:
	Use path separator in place of soft divider.
:param bool use_shortened_path:
	Use path from shortened_path ``--renderer-arg`` argument. If this argument
	is present ``shorten_home`` argument is ignored.
:param bool shorten_home:
	Shorten home directory to ``~``.
:param str ellipsis:
	Specifies what to use in place of omitted directories. Use None to not
	show this subsegment at all.

Divider highlight group used: ``cwd:divider``.

Highlight groups used: ``cwd:current_folder`` or ``cwd``. It is recommended to define all highlight groups.
''')



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import logging

from collections import defaultdict
from itertools import chain
from functools import partial

from powerline import generate_config_finder, get_config_paths, load_config
from powerline.segments.vim import vim_modes
from powerline.lib.dict import mergedicts_copy
from powerline.lib.config import ConfigLoader
from powerline.lib.unicode import unicode
from powerline.lib.path import join
from powerline.lint.markedjson import load
from powerline.lint.markedjson.error import echoerr, EchoErr, MarkedError
from powerline.lint.checks import (check_matcher_func, check_ext, check_config, check_top_theme,
                                   check_color, check_translated_group_name, check_group,
                                   check_segment_module, check_exinclude_function, type_keys,
                                   check_segment_function, check_args, get_one_segment_function,
                                   check_highlight_groups, check_highlight_group, check_full_segment_data,
                                   get_all_possible_functions, check_segment_data_key, register_common_name,
                                   highlight_group_spec, check_log_file_level, check_logging_handler)
from powerline.lint.spec import Spec
from powerline.lint.context import Context


def open_file(path):
	return open(path, 'rb')


def generate_json_config_loader(lhadproblem):
	def load_json_config(config_file_path, load=load, open_file=open_file):
		with open_file(config_file_path) as config_file_fp:
			r, hadproblem = load(config_file_fp)
			if hadproblem:
				lhadproblem[0] = True
			return r
	return load_json_config


function_name_re = r'^(\w+\.)*[a-zA-Z_]\w*$'


divider_spec = Spec().printable().len(
	'le', 3, (lambda value: 'Divider {0!r} is too large!'.format(value))).copy
ext_theme_spec = Spec().type(unicode).func(lambda *args: check_config('themes', *args)).copy
top_theme_spec = Spec().type(unicode).func(check_top_theme).copy
ext_spec = Spec(
	colorscheme=Spec().type(unicode).func(
		(lambda *args: check_config('colorschemes', *args))
	),
	theme=ext_theme_spec(),
	top_theme=top_theme_spec().optional(),
).copy
gen_components_spec = (lambda *components: Spec().list(Spec().type(unicode).oneof(set(components))))
log_level_spec = Spec().re('^[A-Z]+$').func(
	(lambda value, *args: (True, True, not hasattr(logging, value))),
	(lambda value: 'unknown debugging level {0}'.format(value))
).copy
log_format_spec = Spec().type(unicode).copy
main_spec = (Spec(
	common=Spec(
		default_top_theme=top_theme_spec().optional(),
		term_truecolor=Spec().type(bool).optional(),
		term_escape_style=Spec().type(unicode).oneof(set(('auto', 'xterm', 'fbterm'))).optional(),
		# Python is capable of loading from zip archives. Thus checking path 
		# only for existence of the path, not for it being a directory
		paths=Spec().list(
			(lambda value, *args: (True, True, not os.path.exists(os.path.expanduser(value.value)))),
			(lambda value: 'path does not exist: {0}'.format(value))
		).optional(),
		log_file=Spec().either(
			Spec().type(unicode).func(
				(
					lambda value, *args: (
						True,
						True,
						not os.path.isdir(os.path.dirname(os.path.expanduser(value)))
					)
				),
				(lambda value: 'directory does not exist: {0}'.format(os.path.dirname(value)))
			),
			Spec().list(Spec().either(
				Spec().type(unicode, type(None)),
				Spec().tuple(
					Spec().re(function_name_re).func(check_logging_handler),
					Spec().tuple(
						Spec().type(list).optional(),
						Spec().type(dict).optional(),
					),
					log_level_spec().func(check_log_file_level).optional(),
					log_format_spec().optional(),
				),
			))
		).optional(),
		log_level=log_level_spec().optional(),
		log_format=log_format_spec().optional(),
		interval=Spec().either(Spec().cmp('gt', 0.0), Spec().type(type(None))).optional(),
		reload_config=Spec().type(bool).optional(),
		watcher=Spec().type(unicode).oneof(set(('auto', 'inotify', 'stat'))).optional(),
	).context_message('Error while loading common configuration (key {key})'),
	ext=Spec(
		vim=ext_spec().update(
			components=gen_components_spec('statusline', 'tabline').optional(),
			local_themes=Spec(
				__tabline__=ext_theme_spec(),
			).unknown_spec(
				Spec().re(function_name_re).func(partial(check_matcher_func, 'vim')),
				ext_theme_spec()
			),
		).optional(),
		ipython=ext_spec().update(
			local_themes=Spec(
				in2=ext_theme_spec(),
				out=ext_theme_spec(),
				rewrite=ext_theme_spec(),
			),
		).optional(),
		shell=ext_spec().update(
			components=gen_components_spec('tmux', 'prompt').optional(),
			local_themes=Spec(
				continuation=ext_theme_spec(),
				select=ext_theme_spec(),
			),
		).optional(),
		wm=ext_spec().update(
			local_themes=Spec().unknown_spec(
				Spec().re('^[0-9A-Za-z-]+$'),
				ext_theme_spec()
			).optional(),
			update_interval=Spec().cmp('gt', 0.0).optional(),
		).optional(),
	).unknown_spec(
		check_ext,
		ext_spec(),
	).context_message('Error while loading extensions configuration (key {key})'),
).context_message('Error while loading main configuration'))

term_color_spec = Spec().unsigned().cmp('le', 255).copy
true_color_spec = Spec().re(
	'^[0-9a-fA-F]{6}$',
	(lambda value: '"{0}" is not a six-digit hexadecimal unsigned integer written as a string'.format(value))
).copy
colors_spec = (Spec(
	colors=Spec().unknown_spec(
		Spec().ident(),
		Spec().either(
			Spec().tuple(term_color_spec(), true_color_spec()),
			term_color_spec()
		)
	).context_message('Error while checking colors (key {key})'),
	gradients=Spec().unknown_spec(
		Spec().ident(),
		Spec().tuple(
			Spec().len('gt', 1).list(term_color_spec()),
			Spec().len('gt', 1).list(true_color_spec()).optional(),
		)
	).context_message('Error while checking gradients (key {key})'),
).context_message('Error while loading colors configuration'))


color_spec = Spec().type(unicode).func(check_color).copy
name_spec = Spec().type(unicode).len('gt', 0).optional().copy
group_name_spec = Spec().ident().copy
group_spec = Spec().either(Spec(
	fg=color_spec(),
	bg=color_spec(),
	attrs=Spec().list(Spec().type(unicode).oneof(set(('bold', 'italic', 'underline')))),
), group_name_spec().func(check_group)).copy
groups_spec = Spec().unknown_spec(
	group_name_spec(),
	group_spec(),
).context_message('Error while loading groups (key {key})').copy
colorscheme_spec = (Spec(
	name=name_spec(),
	groups=groups_spec(),
).context_message('Error while loading coloscheme'))
mode_translations_value_spec = Spec(
	colors=Spec().unknown_spec(
		color_spec(),
		color_spec(),
	).optional(),
	groups=Spec().unknown_spec(
		group_name_spec().func(check_translated_group_name),
		group_spec(),
	).optional(),
).copy
top_colorscheme_spec = (Spec(
	name=name_spec(),
	groups=groups_spec(),
	mode_translations=Spec().unknown_spec(
		Spec().type(unicode),
		mode_translations_value_spec(),
	).optional().context_message('Error while loading mode translations (key {key})').optional(),
).context_message('Error while loading top-level coloscheme'))
vim_mode_spec = Spec().oneof(set(list(vim_modes) + ['nc', 'tab_nc', 'buf_nc'])).copy
vim_colorscheme_spec = (Spec(
	name=name_spec(),
	groups=groups_spec(),
	mode_translations=Spec().unknown_spec(
		vim_mode_spec(),
		mode_translations_value_spec(),
	).optional().context_message('Error while loading mode translations (key {key})'),
).context_message('Error while loading vim colorscheme'))
shell_mode_spec = Spec().re(r'^(?:[\w\-]+|\.safe)$').copy
shell_colorscheme_spec = (Spec(
	name=name_spec(),
	groups=groups_spec(),
	mode_translations=Spec().unknown_spec(
		shell_mode_spec(),
		mode_translations_value_spec(),
	).optional().context_message('Error while loading mode translations (key {key})'),
).context_message('Error while loading shell colorscheme'))


args_spec = Spec(
	pl=Spec().error('pl object must be set by powerline').optional(),
	segment_info=Spec().error('Segment info dictionary must be set by powerline').optional(),
).unknown_spec(Spec(), Spec()).optional().copy
segment_module_spec = Spec().type(unicode).func(check_segment_module).optional().copy
exinclude_spec = Spec().re(function_name_re).func(check_exinclude_function).copy
segment_spec_base = Spec(
	name=Spec().re(r'^[a-zA-Z_]\w*$').optional(),
	function=Spec().re(function_name_re).func(check_segment_function).optional(),
	exclude_modes=Spec().list(vim_mode_spec()).optional(),
	include_modes=Spec().list(vim_mode_spec()).optional(),
	exclude_function=exinclude_spec().optional(),
	include_function=exinclude_spec().optional(),
	draw_hard_divider=Spec().type(bool).optional(),
	draw_soft_divider=Spec().type(bool).optional(),
	draw_inner_divider=Spec().type(bool).optional(),
	display=Spec().type(bool).optional(),
	module=segment_module_spec(),
	priority=Spec().type(int, float, type(None)).optional(),
	after=Spec().printable().optional(),
	before=Spec().printable().optional(),
	width=Spec().either(Spec().unsigned(), Spec().cmp('eq', 'auto')).optional(),
	align=Spec().oneof(set('lr')).optional(),
	args=args_spec().func(lambda *args, **kwargs: check_args(get_one_segment_function, *args, **kwargs)),
	contents=Spec().printable().optional(),
	highlight_groups=Spec().list(
		highlight_group_spec().re(
			'^(?:(?!:divider$).)+$',
			(lambda value: 'it is recommended that only divider highlight group names end with ":divider"')
		)
	).func(check_highlight_groups).optional(),
	divider_highlight_group=highlight_group_spec().func(check_highlight_group).re(
		':divider$',
		(lambda value: 'it is recommended that divider highlight group names end with ":divider"')
	).optional(),
).func(check_full_segment_data).copy
subsegment_spec = segment_spec_base().update(
	type=Spec().oneof(set((key for key in type_keys if key != 'segment_list'))).optional(),
)
segment_spec = segment_spec_base().update(
	type=Spec().oneof(type_keys).optional(),
	segments=Spec().optional().list(subsegment_spec),
)
segments_spec = Spec().optional().list(segment_spec).copy
segdict_spec = Spec(
	left=segments_spec().context_message('Error while loading segments from left side (key {key})'),
	right=segments_spec().context_message('Error while loading segments from right side (key {key})'),
).func(
	(lambda value, *args: (True, True, not (('left' in value) or ('right' in value)))),
	(lambda value: 'segments dictionary must contain either left, right or both keys')
).context_message('Error while loading segments (key {key})').copy
divside_spec = Spec(
	hard=divider_spec(),
	soft=divider_spec(),
).copy
segment_data_value_spec = Spec(
	after=Spec().printable().optional(),
	before=Spec().printable().optional(),
	display=Spec().type(bool).optional(),
	args=args_spec().func(lambda *args, **kwargs: check_args(get_all_possible_functions, *args, **kwargs)),
	contents=Spec().printable().optional(),
).copy
dividers_spec = Spec(
	left=divside_spec(),
	right=divside_spec(),
).copy
spaces_spec = Spec().unsigned().cmp(
	'le', 2, (lambda value: 'Are you sure you need such a big ({0}) number of spaces?'.format(value))
).copy
common_theme_spec = Spec(
	default_module=segment_module_spec().optional(),
	cursor_space=Spec().type(int, float).cmp('le', 100).cmp('gt', 0).optional(),
	cursor_columns=Spec().type(int).cmp('gt', 0).optional(),
).context_message('Error while loading theme').copy
top_theme_spec = common_theme_spec().update(
	dividers=dividers_spec(),
	spaces=spaces_spec(),
	use_non_breaking_spaces=Spec().type(bool).optional(),
	segment_data=Spec().unknown_spec(
		Spec().func(check_segment_data_key),
		segment_data_value_spec(),
	).optional().context_message('Error while loading segment data (key {key})'),
)
main_theme_spec = common_theme_spec().update(
	dividers=dividers_spec().optional(),
	spaces=spaces_spec().optional(),
	segment_data=Spec().unknown_spec(
		Spec().func(check_segment_data_key),
		segment_data_value_spec(),
	).optional().context_message('Error while loading segment data (key {key})'),
)
theme_spec = common_theme_spec().update(
	dividers=dividers_spec().optional(),
	spaces=spaces_spec().optional(),
	segment_data=Spec().unknown_spec(
		Spec().func(check_segment_data_key),
		segment_data_value_spec(),
	).optional().context_message('Error while loading segment data (key {key})'),
	segments=segdict_spec().update(above=Spec().list(segdict_spec()).optional()),
)


def register_common_names():
	register_common_name('player', 'powerline.segments.common.players', '_player')


def load_json_file(path):
	with open_file(path) as F:
		try:
			config, hadproblem = load(F)
		except MarkedError as e:
			return True, None, str(e)
		else:
			return hadproblem, config, None


def updated_with_config(d):
	hadproblem, config, error = load_json_file(d['path'])
	d.update(
		hadproblem=hadproblem,
		config=config,
		error=error,
	)
	return d


def find_all_ext_config_files(search_paths, subdir):
	for config_root in search_paths:
		top_config_subpath = join(config_root, subdir)
		if not os.path.isdir(top_config_subpath):
			if os.path.exists(top_config_subpath):
				yield {
					'error': 'Path {0} is not a directory'.format(top_config_subpath),
					'path': top_config_subpath,
				}
			continue
		for ext_name in os.listdir(top_config_subpath):
			ext_path = os.path.join(top_config_subpath, ext_name)
			if not os.path.isdir(ext_path):
				if ext_name.endswith('.json') and os.path.isfile(ext_path):
					yield updated_with_config({
						'error': False,
						'path': ext_path,
						'name': ext_name[:-5],
						'ext': None,
						'type': 'top_' + subdir,
					})
				else:
					yield {
						'error': 'Path {0} is not a directory or configuration file'.format(ext_path),
						'path': ext_path,
					}
				continue
			for config_file_name in os.listdir(ext_path):
				config_file_path = os.path.join(ext_path, config_file_name)
				if config_file_name.endswith('.json') and os.path.isfile(config_file_path):
					yield updated_with_config({
						'error': False,
						'path': config_file_path,
						'name': config_file_name[:-5],
						'ext': ext_name,
						'type': subdir,
					})
				else:
					yield {
						'error': 'Path {0} is not a configuration file'.format(config_file_path),
						'path': config_file_path,
					}


def dict2(d):
	return defaultdict(dict, ((k, dict(v)) for k, v in d.items()))


def check(paths=None, debug=False, echoerr=echoerr, require_ext=None):
	'''Check configuration sanity

	:param list paths:
		Paths from which configuration should be loaded.
	:param bool debug:
		Determines whether some information useful for debugging linter should 
		be output.
	:param function echoerr:
		Function that will be used to echo the error(s). Should accept four 
		optional keyword parameters: ``problem`` and ``problem_mark``, and 
		``context`` and ``context_mark``.
	:param str require_ext:
		Require configuration for some extension to be present.

	:return:
		``False`` if user configuration seems to be completely sane and ``True`` 
		if some problems were found.
	'''
	hadproblem = False

	register_common_names()
	search_paths = paths or get_config_paths()
	find_config_files = generate_config_finder(lambda: search_paths)

	logger = logging.getLogger('powerline-lint')
	logger.setLevel(logging.DEBUG if debug else logging.ERROR)
	logger.addHandler(logging.StreamHandler())

	ee = EchoErr(echoerr, logger)

	if require_ext:
		used_main_spec = main_spec.copy()
		try:
			used_main_spec['ext'][require_ext].required()
		except KeyError:
			used_main_spec['ext'][require_ext] = ext_spec()
	else:
		used_main_spec = main_spec

	lhadproblem = [False]
	load_json_config = generate_json_config_loader(lhadproblem)

	config_loader = ConfigLoader(run_once=True, load=load_json_config)

	lists = {
		'colorschemes': set(),
		'themes': set(),
		'exts': set(),
	}
	found_dir = {
		'themes': False,
		'colorschemes': False,
	}
	config_paths = defaultdict(lambda: defaultdict(dict))
	loaded_configs = defaultdict(lambda: defaultdict(dict))
	for d in chain(
		find_all_ext_config_files(search_paths, 'colorschemes'),
		find_all_ext_config_files(search_paths, 'themes'),
	):
		if d['error']:
			hadproblem = True
			ee(problem=d['error'])
			continue
		if d['hadproblem']:
			hadproblem = True
		if d['ext']:
			found_dir[d['type']] = True
			lists['exts'].add(d['ext'])
			if d['name'] == '__main__':
				pass
			elif d['name'].startswith('__') or d['name'].endswith('__'):
				hadproblem = True
				ee(problem='File name is not supposed to start or end with “__”: {0}'.format(
					d['path']))
			else:
				lists[d['type']].add(d['name'])
			config_paths[d['type']][d['ext']][d['name']] = d['path']
			loaded_configs[d['type']][d['ext']][d['name']] = d['config']
		else:
			config_paths[d['type']][d['name']] = d['path']
			loaded_configs[d['type']][d['name']] = d['config']

	for typ in ('themes', 'colorschemes'):
		if not found_dir[typ]:
			hadproblem = True
			ee(problem='Subdirectory {0} was not found in paths {1}'.format(typ, ', '.join(search_paths)))

	diff = set(config_paths['colorschemes']) - set(config_paths['themes'])
	if diff:
		hadproblem = True
		for ext in diff:
			typ = 'colorschemes' if ext in config_paths['themes'] else 'themes'
			if not config_paths['top_' + typ] or typ == 'themes':
				ee(problem='{0} extension {1} not present in {2}'.format(
					ext,
					'configuration' if (
						ext in loaded_configs['themes'] and ext in loaded_configs['colorschemes']
					) else 'directory',
					typ,
				))

	try:
		main_config = load_config('config', find_config_files, config_loader)
	except IOError:
		main_config = {}
		ee(problem='Configuration file not found: config.json')
		hadproblem = True
	except MarkedError as e:
		main_config = {}
		ee(problem=str(e))
		hadproblem = True
	else:
		if used_main_spec.match(
			main_config,
			data={'configs': config_paths, 'lists': lists},
			context=Context(main_config),
			echoerr=ee
		)[1]:
			hadproblem = True

	import_paths = [os.path.expanduser(path) for path in main_config.get('common', {}).get('paths', [])]

	try:
		colors_config = load_config('colors', find_config_files, config_loader)
	except IOError:
		colors_config = {}
		ee(problem='Configuration file not found: colors.json')
		hadproblem = True
	except MarkedError as e:
		colors_config = {}
		ee(problem=str(e))
		hadproblem = True
	else:
		if colors_spec.match(colors_config, context=Context(colors_config), echoerr=ee)[1]:
			hadproblem = True

	if lhadproblem[0]:
		hadproblem = True

	top_colorscheme_configs = dict(loaded_configs['top_colorschemes'])
	data = {
		'ext': None,
		'top_colorscheme_configs': top_colorscheme_configs,
		'ext_colorscheme_configs': {},
		'colors_config': colors_config
	}
	for colorscheme, config in loaded_configs['top_colorschemes'].items():
		data['colorscheme'] = colorscheme
		if top_colorscheme_spec.match(config, context=Context(config), data=data, echoerr=ee)[1]:
			hadproblem = True

	ext_colorscheme_configs = dict2(loaded_configs['colorschemes'])
	for ext, econfigs in ext_colorscheme_configs.items():
		data = {
			'ext': ext,
			'top_colorscheme_configs': top_colorscheme_configs,
			'ext_colorscheme_configs': ext_colorscheme_configs,
			'colors_config': colors_config,
		}
		for colorscheme, config in econfigs.items():
			data['colorscheme'] = colorscheme
			if ext == 'vim':
				spec = vim_colorscheme_spec
			elif ext == 'shell':
				spec = shell_colorscheme_spec
			else:
				spec = colorscheme_spec
			if spec.match(config, context=Context(config), data=data, echoerr=ee)[1]:
				hadproblem = True

	colorscheme_configs = {}
	for ext in lists['exts']:
		colorscheme_configs[ext] = {}
		for colorscheme in lists['colorschemes']:
			econfigs = ext_colorscheme_configs[ext]
			ecconfigs = econfigs.get(colorscheme)
			mconfigs = (
				top_colorscheme_configs.get(colorscheme),
				econfigs.get('__main__'),
				ecconfigs,
			)
			if not (mconfigs[0] or mconfigs[2]):
				continue
			config = None
			for mconfig in mconfigs:
				if not mconfig:
					continue
				if config:
					config = mergedicts_copy(config, mconfig)
				else:
					config = mconfig
			colorscheme_configs[ext][colorscheme] = config

	theme_configs = dict2(loaded_configs['themes'])
	top_theme_configs = dict(loaded_configs['top_themes'])
	for ext, configs in theme_configs.items():
		data = {
			'ext': ext,
			'colorscheme_configs': colorscheme_configs,
			'import_paths': import_paths,
			'main_config': main_config,
			'top_themes': top_theme_configs,
			'ext_theme_configs': configs,
			'colors_config': colors_config
		}
		for theme, config in configs.items():
			data['theme'] = theme
			if theme == '__main__':
				data['theme_type'] = 'main'
				spec = main_theme_spec
			else:
				data['theme_type'] = 'regular'
				spec = theme_spec
			if spec.match(config, context=Context(config), data=data, echoerr=ee)[1]:
				hadproblem = True

	for top_theme, config in top_theme_configs.items():
		data = {
			'ext': None,
			'colorscheme_configs': colorscheme_configs,
			'import_paths': import_paths,
			'main_config': main_config,
			'theme_configs': theme_configs,
			'ext_theme_configs': None,
			'colors_config': colors_config
		}
		data['theme_type'] = 'top'
		data['theme'] = top_theme
		if top_theme_spec.match(config, context=Context(config), data=data, echoerr=ee)[1]:
			hadproblem = True

	return hadproblem



# vim:fileencoding=utf-8:noet
# WARNING: using unicode_literals causes errors in argparse
from __future__ import (division, absolute_import, print_function)

import argparse
import sys

from itertools import chain

from powerline.lib.overrides import parsedotval, parse_override_var
from powerline.lib.dict import mergeargs
from powerline.lib.encoding import get_preferred_arguments_encoding
from powerline.lib.unicode import u, unicode
from powerline.bindings.wm import wm_threads


if sys.version_info < (3,):
	encoding = get_preferred_arguments_encoding()

	def arg_to_unicode(s):
		return unicode(s, encoding, 'replace') if not isinstance(s, unicode) else s  # NOQA
else:
	def arg_to_unicode(s):
		return s


def finish_args(parser, environ, args, is_daemon=False):
	'''Do some final transformations

	Transforms ``*_override`` arguments into dictionaries, adding overrides from 
	environment variables. Transforms ``renderer_arg`` argument into dictionary 
	as well, but only if it is true.

	:param dict environ:
		Environment from which additional overrides should be taken from.
	:param args:
		Arguments object returned by 
		:py:meth:`argparse.ArgumentParser.parse_args`. Will be modified 
		in-place.

	:return: Object received as second (``args``) argument.
	'''
	args.config_override = mergeargs(chain(
		parse_override_var(environ.get('POWERLINE_CONFIG_OVERRIDES', '')),
		(parsedotval(v) for v in args.config_override or ()),
	))
	args.theme_override = mergeargs(chain(
		parse_override_var(environ.get('POWERLINE_THEME_OVERRIDES', '')),
		(parsedotval(v) for v in args.theme_override or ()),
	))
	if args.renderer_arg:
		args.renderer_arg = mergeargs((parsedotval(v) for v in args.renderer_arg), remove=True)
		if 'pane_id' in args.renderer_arg:
			if isinstance(args.renderer_arg['pane_id'], (bytes, unicode)):
				try:
					args.renderer_arg['pane_id'] = int(args.renderer_arg['pane_id'].lstrip(' %'))
				except ValueError:
					pass
			if 'client_id' not in args.renderer_arg:
				args.renderer_arg['client_id'] = args.renderer_arg['pane_id']
	args.config_path = (
		[path for path in environ.get('POWERLINE_CONFIG_PATHS', '').split(':') if path]
		+ (args.config_path or [])
	)
	if args.ext[0].startswith('wm.'):
		if not is_daemon:
			parser.error('WM bindings must be used with daemon only')
		elif args.ext[0][3:] not in wm_threads:
			parser.error('WM binding not found')
	elif not args.side:
		parser.error('expected one argument')
	return args


def int_or_sig(s):
	if s.startswith('sig'):
		return u(s)
	else:
		return int(s)


def get_argparser(ArgumentParser=argparse.ArgumentParser):
	parser = ArgumentParser(description='Powerline prompt and statusline script.')
	parser.add_argument(
		'ext', nargs=1,
		help='Extension: application for which powerline command is launched '
		     '(usually `shell\' or `tmux\'). Also supports `wm.\' extensions: '
		     + ', '.join(('`wm.' + key + '\'' for key in wm_threads.keys())) + '.'
	)
	parser.add_argument(
		'side', nargs='?', choices=('left', 'right', 'above', 'aboveleft'),
		help='Side: `left\' and `right\' represent left and right side '
		     'respectively, `above\' emits lines that are supposed to be printed '
		     'just above the prompt and `aboveleft\' is like concatenating '
		     '`above\' with `left\' with the exception that only one Python '
		     'instance is used in this case. May be omitted for `wm.*\' extensions.'
	)
	parser.add_argument(
		'-r', '--renderer-module', metavar='MODULE', type=str,
		help='Renderer module. Usually something like `.bash\' or `.zsh\' '
		     '(with leading dot) which is `powerline.renderers.{ext}{MODULE}\', '
		     'may also be full module name (must contain at least one dot or '
		     'end with a dot in case it is top-level module) or '
		     '`powerline.renderers\' submodule (in case there are no dots).'
	)
	parser.add_argument(
		'-w', '--width', type=int,
		help='Maximum prompt with. Triggers truncation of some segments.'
	)
	parser.add_argument(
		'--last-exit-code', metavar='INT', type=int_or_sig,
		help='Last exit code.'
	)
	parser.add_argument(
		'--last-pipe-status', metavar='LIST', default='',
		type=lambda s: [int_or_sig(status) for status in s.split()],
		help='Like above, but is supposed to contain space-separated array '
		     'of statuses, representing exit statuses of commands in one pipe.'
	)
	parser.add_argument(
		'--jobnum', metavar='INT', type=int,
		help='Number of jobs.'
	)
	parser.add_argument(
		'-c', '--config-override', metavar='KEY.KEY=VALUE', type=arg_to_unicode,
		action='append',
		help='Configuration overrides for `config.json\'. Is translated to a '
		     'dictionary and merged with the dictionary obtained from actual '
		     'JSON configuration: KEY.KEY=VALUE is translated to '
		     '`{"KEY": {"KEY": VALUE}}\' and then merged recursively. '
		     'VALUE may be any JSON value, values that are not '
		     '`null\', `true\', `false\', start with digit, `{\', `[\' '
		     'are treated like strings. If VALUE is omitted '
		     'then corresponding key is removed.'
	)
	parser.add_argument(
		'-t', '--theme-override', metavar='THEME.KEY.KEY=VALUE', type=arg_to_unicode,
		action='append',
		help='Like above, but theme-specific. THEME should point to '
		     'an existing and used theme to have any effect, but it is fine '
		     'to use any theme here.'
	)
	parser.add_argument(
		'-R', '--renderer-arg',
		metavar='KEY=VAL', type=arg_to_unicode, action='append',
		help='Like above, but provides argument for renderer. Is supposed '
		     'to be used only by shell bindings to provide various data like '
		     'last-exit-code or last-pipe-status (they are not using '
		     '`--renderer-arg\' for historical reasons: `--renderer-arg\' '
		     'was added later).'
	)
	parser.add_argument(
		'-p', '--config-path', action='append', metavar='PATH',
		help='Path to configuration directory. If it is present then '
		     'configuration files will only be sought in the provided path. '
		     'May be provided multiple times to search in a list of directories.'
	)
	parser.add_argument(
		'--socket', metavar='ADDRESS', type=str,
		help='Socket address to use in daemon clients. Is always UNIX domain '
		     'socket on linux and file socket on Mac OS X. Not used here, '
		     'present only for compatibility with other powerline clients. '
		     'This argument must always be the first one and be in a form '
		     '`--socket ADDRESS\': no `=\' or short form allowed '
		     '(in other powerline clients, not here).'
	)
	return parser


def write_output(args, powerline, segment_info, write):
	if args.renderer_arg:
		segment_info.update(args.renderer_arg)
	if args.side.startswith('above'):
		for line in powerline.render_above_lines(
			width=args.width,
			segment_info=segment_info,
			mode=segment_info.get('mode', None),
		):
			if line:
				write(line + '\n')
		args.side = args.side[len('above'):]

	if args.side:
		rendered = powerline.render(
			width=args.width,
			side=args.side,
			segment_info=segment_info,
			mode=segment_info.get('mode', None),
		)
		write(rendered)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys

from pkgutil import extend_path
from types import MethodType


__path__ = extend_path(__path__, __name__)


class Segment(object):
	'''Base class for any segment that is not a function

	Required for powerline.lint.inspect to work properly: it defines methods for 
	omitting existing or adding new arguments.

	.. note::
		Until python-3.4 ``inspect.getargspec`` does not support querying 
		callable classes for arguments of their ``__call__`` method, requiring 
		to use this method directly (i.e. before 3.4 you should write 
		``getargspec(obj.__call__)`` in place of ``getargspec(obj)``).
	'''
	if sys.version_info < (3, 4):
		def argspecobjs(self):
			yield '__call__', self.__call__
	else:
		def argspecobjs(self):
			yield '__call__', self

	argspecobjs.__doc__ = (
		'''Return a list of valid arguments for inspect.getargspec

		Used to determine function arguments.
		'''
	)

	def omitted_args(self, name, method):
		'''List arguments which should be omitted

		Returns a tuple with indexes of omitted arguments.

		.. note::``segment_info``, ``create_watcher`` and ``pl`` will be omitted 
			regardless of the below return (for ``segment_info`` and 
			``create_watcher``: only if object was marked to require segment 
			info or filesystem watcher).
		'''
		if isinstance(self.__call__, MethodType):
			return (0,)
		else:
			return ()

	@staticmethod
	def additional_args():
		'''Returns a list of (additional argument name[, default value]) tuples.
		'''
		return ()


def with_docstring(instance, doc):
	instance.__doc__ = doc
	return instance



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderers.shell import ShellRenderer
from powerline.renderers.shell.readline import ReadlineRenderer
from powerline.renderers.ipython import IPythonRenderer


class IPythonPre50Renderer(IPythonRenderer, ShellRenderer):
	'''Powerline ipython segment renderer for pre-5.0 IPython versions.'''
	def render(self, **kwargs):
		# XXX super(ShellRenderer), *not* super(IPythonPre50Renderer)
		return super(ShellRenderer, self).render(**kwargs)

	def do_render(self, segment_info, **kwargs):
		segment_info.update(client_id='ipython')
		return super(IPythonPre50Renderer, self).do_render(
			segment_info=segment_info,
			**kwargs
		)


class IPythonPromptRenderer(IPythonPre50Renderer, ReadlineRenderer):
	'''Powerline ipython prompt (in and in2) renderer'''
	pass


class IPythonNonPromptRenderer(IPythonPre50Renderer):
	'''Powerline ipython non-prompt (out and rewrite) renderer'''
	pass


class RendererProxy(object):
	'''Powerline IPython renderer proxy which chooses appropriate renderer

	Instantiates two renderer objects: one will be used for prompts and the 
	other for non-prompts.
	'''
	def __init__(self, **kwargs):
		old_widths = {}
		self.non_prompt_renderer = IPythonNonPromptRenderer(old_widths=old_widths, **kwargs)
		self.prompt_renderer = IPythonPromptRenderer(old_widths=old_widths, **kwargs)

	def render_above_lines(self, *args, **kwargs):
		return self.non_prompt_renderer.render_above_lines(*args, **kwargs)

	def render(self, is_prompt, *args, **kwargs):
		return (self.prompt_renderer if is_prompt else self.non_prompt_renderer).render(
			*args, **kwargs)

	def shutdown(self, *args, **kwargs):
		self.prompt_renderer.shutdown(*args, **kwargs)
		self.non_prompt_renderer.shutdown(*args, **kwargs)


renderer = RendererProxy



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from string import hexdigits

from powerline.lint.markedjson.error import MarkedError
from powerline.lint.markedjson import tokens
from powerline.lib.unicode import unicode, unichr, surrogate_pair_to_character


hexdigits_set = set(hexdigits)


# Scanner produces tokens of the following types:
# STREAM-START
# STREAM-END
# DOCUMENT-START
# DOCUMENT-END
# FLOW-SEQUENCE-START
# FLOW-MAPPING-START
# FLOW-SEQUENCE-END
# FLOW-MAPPING-END
# FLOW-ENTRY
# KEY
# VALUE
# SCALAR(value, plain, style)
#
# Read comments in the Scanner code for more details.


class ScannerError(MarkedError):
	pass


class SimpleKey:
	# See below simple keys treatment.
	def __init__(self, token_number, index, line, column, mark):
		self.token_number = token_number
		self.index = index
		self.line = line
		self.column = column
		self.mark = mark


class Scanner:
	def __init__(self):
		'''Initialize the scanner.'''
		# It is assumed that Scanner and Reader will have a common descendant.
		# Reader do the dirty work of checking for BOM and converting the
		# input data to Unicode. It also adds NUL to the end.
		#
		# Reader supports the following methods
		# 	self.peek(i=0)		 # peek the next i-th character
		# 	self.prefix(l=1)	 # peek the next l characters
		# 	self.forward(l=1)	 # read the next l characters and move the pointer.

		# Had we reached the end of the stream?
		self.done = False

		# The number of unclosed '{' and '['. `flow_level == 0` means block
		# context.
		self.flow_level = 0

		# List of processed tokens that are not yet emitted.
		self.tokens = []

		# Add the STREAM-START token.
		self.fetch_stream_start()

		# Number of tokens that were emitted through the `get_token` method.
		self.tokens_taken = 0

		# Variables related to simple keys treatment.

		# A simple key is a key that is not denoted by the '?' indicator.
		# We emit the KEY token before all keys, so when we find a potential
		# simple key, we try to locate the corresponding ':' indicator.
		# Simple keys should be limited to a single line.

		# Can a simple key start at the current position? A simple key may
		# start:
		# - after '{', '[', ',' (in the flow context),
		self.allow_simple_key = False

		# Keep track of possible simple keys. This is a dictionary. The key
		# is `flow_level`; there can be no more that one possible simple key
		# for each level. The value is a SimpleKey record:
		# 	(token_number, index, line, column, mark)
		# A simple key may start with SCALAR(flow), '[', or '{' tokens.
		self.possible_simple_keys = {}

	# Public methods.

	def check_token(self, *choices):
		# Check if the next token is one of the given types.
		while self.need_more_tokens():
			self.fetch_more_tokens()
		if self.tokens:
			if not choices:
				return True
			for choice in choices:
				if isinstance(self.tokens[0], choice):
					return True
		return False

	def peek_token(self):
		# Return the next token, but do not delete if from the queue.
		while self.need_more_tokens():
			self.fetch_more_tokens()
		if self.tokens:
			return self.tokens[0]

	def get_token(self):
		# Return the next token.
		while self.need_more_tokens():
			self.fetch_more_tokens()
		if self.tokens:
			self.tokens_taken += 1
			return self.tokens.pop(0)

	# Private methods.

	def need_more_tokens(self):
		if self.done:
			return False
		if not self.tokens:
			return True
		# The current token may be a potential simple key, so we
		# need to look further.
		self.stale_possible_simple_keys()
		if self.next_possible_simple_key() == self.tokens_taken:
			return True

	def fetch_more_tokens(self):

		# Eat whitespaces and comments until we reach the next token.
		self.scan_to_next_token()

		# Remove obsolete possible simple keys.
		self.stale_possible_simple_keys()

		# Peek the next character.
		ch = self.peek()

		# Is it the end of stream?
		if ch == '\0':
			return self.fetch_stream_end()

		# Note: the order of the following checks is NOT significant.

		# Is it the flow sequence start indicator?
		if ch == '[':
			return self.fetch_flow_sequence_start()

		# Is it the flow mapping start indicator?
		if ch == '{':
			return self.fetch_flow_mapping_start()

		# Is it the flow sequence end indicator?
		if ch == ']':
			return self.fetch_flow_sequence_end()

		# Is it the flow mapping end indicator?
		if ch == '}':
			return self.fetch_flow_mapping_end()

		# Is it the flow entry indicator?
		if ch == ',':
			return self.fetch_flow_entry()

		# Is it the value indicator?
		if ch == ':' and self.flow_level:
			return self.fetch_value()

		# Is it a double quoted scalar?
		if ch == '"':
			return self.fetch_double()

		# It must be a plain scalar then.
		if self.check_plain():
			return self.fetch_plain()

		# No? It’s an error. Let’s produce a nice error message.
		raise ScannerError(
			'while scanning for the next token', None,
			'found character %r that cannot start any token' % ch,
			self.get_mark()
		)

	# Simple keys treatment.

	def next_possible_simple_key(self):
		# Return the number of the nearest possible simple key. Actually we
		# don’t need to loop through the whole dictionary. We may replace it
		# with the following code:
		# 	if not self.possible_simple_keys:
		# 		return None
		# 	return self.possible_simple_keys[
		# 			min(self.possible_simple_keys.keys())].token_number
		min_token_number = None
		for level in self.possible_simple_keys:
			key = self.possible_simple_keys[level]
			if min_token_number is None or key.token_number < min_token_number:
				min_token_number = key.token_number
		return min_token_number

	def stale_possible_simple_keys(self):
		# Remove entries that are no longer possible simple keys. According to
		# the YAML specification, simple keys
		# - should be limited to a single line,
		# Disabling this procedure will allow simple keys of any length and
		# height (may cause problems if indentation is broken though).
		for level in list(self.possible_simple_keys):
			key = self.possible_simple_keys[level]
			if key.line != self.line:
				del self.possible_simple_keys[level]

	def save_possible_simple_key(self):
		# The next token may start a simple key. We check if it’s possible
		# and save its position. This function is called for
		# 	SCALAR(flow), '[', and '{'.

		# The next token might be a simple key. Let’s save it’s number and
		# position.
		if self.allow_simple_key:
			self.remove_possible_simple_key()
			token_number = self.tokens_taken + len(self.tokens)
			key = SimpleKey(token_number, self.index, self.line, self.column, self.get_mark())
			self.possible_simple_keys[self.flow_level] = key

	def remove_possible_simple_key(self):
		# Remove the saved possible key position at the current flow level.
		if self.flow_level in self.possible_simple_keys:
			del self.possible_simple_keys[self.flow_level]

	# Fetchers.

	def fetch_stream_start(self):
		# We always add STREAM-START as the first token and STREAM-END as the
		# last token.

		# Read the token.
		mark = self.get_mark()

		# Add STREAM-START.
		self.tokens.append(tokens.StreamStartToken(mark, mark, encoding=self.encoding))

	def fetch_stream_end(self):
		# Reset simple keys.
		self.remove_possible_simple_key()
		self.allow_simple_key = False
		self.possible_simple_keys = {}

		# Read the token.
		mark = self.get_mark()

		# Add STREAM-END.
		self.tokens.append(tokens.StreamEndToken(mark, mark))

		# The steam is finished.
		self.done = True

	def fetch_flow_sequence_start(self):
		self.fetch_flow_collection_start(tokens.FlowSequenceStartToken)

	def fetch_flow_mapping_start(self):
		self.fetch_flow_collection_start(tokens.FlowMappingStartToken)

	def fetch_flow_collection_start(self, TokenClass):
		# '[' and '{' may start a simple key.
		self.save_possible_simple_key()

		# Increase the flow level.
		self.flow_level += 1

		# Simple keys are allowed after '[' and '{'.
		self.allow_simple_key = True

		# Add FLOW-SEQUENCE-START or FLOW-MAPPING-START.
		start_mark = self.get_mark()
		self.forward()
		end_mark = self.get_mark()
		self.tokens.append(TokenClass(start_mark, end_mark))

	def fetch_flow_sequence_end(self):
		self.fetch_flow_collection_end(tokens.FlowSequenceEndToken)

	def fetch_flow_mapping_end(self):
		self.fetch_flow_collection_end(tokens.FlowMappingEndToken)

	def fetch_flow_collection_end(self, TokenClass):
		# Reset possible simple key on the current level.
		self.remove_possible_simple_key()

		# Decrease the flow level.
		self.flow_level -= 1

		# No simple keys after ']' or '}'.
		self.allow_simple_key = False

		# Add FLOW-SEQUENCE-END or FLOW-MAPPING-END.
		start_mark = self.get_mark()
		self.forward()
		end_mark = self.get_mark()
		self.tokens.append(TokenClass(start_mark, end_mark))

	def fetch_value(self):
		# Do we determine a simple key?
		if self.flow_level in self.possible_simple_keys:

			# Add KEY.
			key = self.possible_simple_keys[self.flow_level]
			del self.possible_simple_keys[self.flow_level]
			self.tokens.insert(key.token_number - self.tokens_taken, tokens.KeyToken(key.mark, key.mark))

			# There cannot be two simple keys one after another.
			self.allow_simple_key = False

		# Add VALUE.
		start_mark = self.get_mark()
		self.forward()
		end_mark = self.get_mark()
		self.tokens.append(tokens.ValueToken(start_mark, end_mark))

	def fetch_flow_entry(self):
		# Simple keys are allowed after ','.
		self.allow_simple_key = True

		# Reset possible simple key on the current level.
		self.remove_possible_simple_key()

		# Add FLOW-ENTRY.
		start_mark = self.get_mark()
		self.forward()
		end_mark = self.get_mark()
		self.tokens.append(tokens.FlowEntryToken(start_mark, end_mark))

	def fetch_double(self):
		# A flow scalar could be a simple key.
		self.save_possible_simple_key()

		# No simple keys after flow scalars.
		self.allow_simple_key = False

		# Scan and add SCALAR.
		self.tokens.append(self.scan_flow_scalar())

	def fetch_plain(self):

		self.save_possible_simple_key()

		# No simple keys after plain scalars.
		self.allow_simple_key = False

		# Scan and add SCALAR. May change `allow_simple_key`.
		self.tokens.append(self.scan_plain())

	# Checkers.

	def check_plain(self):
		return self.peek() in '0123456789-ntf'

	# Scanners.

	def scan_to_next_token(self):
		while self.peek() in ' \t\n':
			self.forward()

	def scan_flow_scalar(self):
		# See the specification for details.
		# Note that we loose indentation rules for quoted scalars. Quoted
		# scalars don’t need to adhere indentation because " and ' clearly
		# mark the beginning and the end of them. Therefore we are less
		# restrictive then the specification requires. We only need to check
		# that document separators are not included in scalars.
		chunks = []
		start_mark = self.get_mark()
		quote = self.peek()
		self.forward()
		chunks.extend(self.scan_flow_scalar_non_spaces(start_mark))
		while self.peek() != quote:
			chunks.extend(self.scan_flow_scalar_spaces(start_mark))
			chunks.extend(self.scan_flow_scalar_non_spaces(start_mark))
		self.forward()
		end_mark = self.get_mark()
		return tokens.ScalarToken(unicode().join(chunks), False, start_mark, end_mark, '"')

	ESCAPE_REPLACEMENTS = {
		'b': '\x08',
		't': '\x09',
		'n': '\x0A',
		'f': '\x0C',
		'r': '\x0D',
		'"': '\"',
		'\\': '\\',
	}

	ESCAPE_CODES = {
		'u': 4,
	}

	def scan_flow_scalar_non_spaces(self, start_mark):
		# See the specification for details.
		chunks = []
		while True:
			length = 0
			while self.peek(length) not in '\"\\\0 \t\n':
				length += 1
			if length:
				chunks.append(self.prefix(length))
				self.forward(length)
			ch = self.peek()
			if ch == '\\':
				self.forward()
				ch = self.peek()
				if ch in self.ESCAPE_REPLACEMENTS:
					chunks.append(self.ESCAPE_REPLACEMENTS[ch])
					self.forward()
				elif ch in self.ESCAPE_CODES:
					length = self.ESCAPE_CODES[ch]
					self.forward()
					for k in range(length):
						if self.peek(k) not in hexdigits:
							raise ScannerError(
								'while scanning a double-quoted scalar', start_mark,
								'expected escape sequence of %d hexdecimal numbers, but found %r' % (
									length, self.peek(k)),
								self.get_mark()
							)
					code = int(self.prefix(length), 16)
					self.forward(length)
					if 0xD800 <= code <= 0xDC00:
						# Start of the surrogate pair
						next_char = self.prefix(6)
						if (
							next_char[0] != '\\'
							or next_char[1] != 'u'
							or not (set(next_char[2:]) < hexdigits_set)
							or not (0xDC00 <= int(next_char[2:], 16) <= 0xDFFF)
						):
							raise ScannerError(
								'while scanning a double-quoted scalar', start_mark,
								'expected escape sequence with the next character in surrogate pair, but found %r' % (
									next_char
								),
								self.get_mark()
							)
						code = surrogate_pair_to_character(code, int(next_char[2:], 16))
						self.forward(6)
					chunks.append(unichr(code))
				else:
					raise ScannerError(
						'while scanning a double-quoted scalar', start_mark,
						('found unknown escape character %r' % ch), self.get_mark()
					)
			else:
				return chunks

	def scan_flow_scalar_spaces(self, start_mark):
		# See the specification for details.
		chunks = []
		length = 0
		while self.peek(length) in ' \t':
			length += 1
		whitespaces = self.prefix(length)
		self.forward(length)
		ch = self.peek()
		if ch == '\0':
			raise ScannerError(
				'while scanning a quoted scalar', start_mark,
				'found unexpected end of stream', self.get_mark()
			)
		elif ch == '\n':
			raise ScannerError(
				'while scanning a quoted scalar', start_mark,
				'found unexpected line end', self.get_mark()
			)
		else:
			chunks.append(whitespaces)
		return chunks

	def scan_plain(self):
		chunks = []
		start_mark = self.get_mark()
		spaces = []
		while True:
			length = 0
			while True:
				if self.peek(length) not in 'eE.0123456789nul-tr+fas':
					break
				length += 1
			if length == 0:
				break
			self.allow_simple_key = False
			chunks.extend(spaces)
			chunks.append(self.prefix(length))
			self.forward(length)
		end_mark = self.get_mark()
		return tokens.ScalarToken(''.join(chunks), True, start_mark, end_mark)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

try:
	import vim
except ImportError:
	vim = object()

from powerline.bindings.vim import vim_command_exists, vim_get_autoload_func
from powerline.theme import requires_segment_info


currenttag = None
tag_cache = {}


@requires_segment_info
def current_tag(segment_info, pl, flags='s'):
	'''Return tag that is near the cursor.

	:param str flags:
		Specifies additional properties of the displayed tag. Supported values:

		* s - display complete signature
		* f - display the full hierarchy of the tag
		* p - display the raw prototype

		More info in the `official documentation`_ (search for 
		“tagbar#currenttag”).

		.. _`official documentation`: https://github.com/majutsushi/tagbar/blob/master/doc/tagbar.txt
	'''
	global currenttag
	global tag_cache
	window_id = segment_info['window_id']
	if segment_info['mode'] == 'nc':
		return tag_cache.get(window_id, (None,))[-1]
	if not currenttag:
		if vim_command_exists('Tagbar'):
			currenttag = vim_get_autoload_func('tagbar#currenttag')
			if not currenttag:
				return None
		else:
			return None
	prev_key, r = tag_cache.get(window_id, (None, None))
	key = (int(vim.eval('b:changedtick')), segment_info['window'].cursor[0])
	if prev_key and key == prev_key:
		return r
	r = currenttag('%s', '', flags)
	tag_cache[window_id] = (key, r)
	return r
